##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Support for owned objects
"""

from Acquisition import aq_base
from Acquisition import aq_get
from Acquisition import aq_inner
from Acquisition import aq_parent
from ExtensionClass import Base
from zope.interface import implementer

from AccessControl import SpecialUsers as SU
from AccessControl.class_init import InitializeClass
from AccessControl.interfaces import IOwned
from AccessControl.Permissions import take_ownership
from AccessControl.Permissions import view_management_screens
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager


UnownableOwner = []


def ownableFilter(self):
    _owner = aq_get(self, '_owner', None, 1)
    return _owner is not UnownableOwner


# Marker to use as a getattr default.
_mark = ownableFilter


@implementer(IOwned)
class Owned(Base):

    security = ClassSecurityInfo()
    security.setPermissionDefault(take_ownership, ('Owner', ))

    @security.protected(view_management_screens)
    def owner_info(self):
        """Get ownership info for display
        """
        owner = self.getOwnerTuple()

        if owner is None or owner is UnownableOwner:
            return owner

        d = {
            'path': '/'.join(owner[0]),
            'id': owner[1],
            'explicit': hasattr(self, '_owner'),
            'userCanChangeOwnershipType':
            getSecurityManager().checkPermission('Take ownership', self),
        }
        return d

    @security.private
    def getOwner(self, info=0,
                 aq_get=aq_get,
                 UnownableOwner=UnownableOwner,
                 getSecurityManager=getSecurityManager,
                 ):
        """Get the owner

        If a true argument is provided, then only the owner path and id are
        returned. Otherwise, the owner object is returned.
        """
        if info:
            import warnings
            warnings.warn('Owned.getOwner(1) is deprecated; '
                          'please use getOwnerTuple() instead.',
                          DeprecationWarning, stacklevel=2)

        owner = aq_get(self, '_owner', None, 1)
        if info or (owner is None):
            return owner

        if owner is UnownableOwner:
            return None

        udb, oid = owner

        root = self.getPhysicalRoot()
        udb = root.unrestrictedTraverse(udb, None)
        if udb is None:
            user = SU.nobody
        else:
            user = udb.getUserById(oid, None)
            if user is None:
                user = SU.nobody
        return user

    @security.private
    def getOwnerTuple(self):
        """Return a tuple, (userdb_path, user_id) for the owner.

        o Ownership can be acquired, but only from the containment path.

        o If unowned, return None.
        """
        return aq_get(self, '_owner', None, 1)

    @security.private
    def getWrappedOwner(self):
        """Get the owner, modestly wrapped in the user folder.

        o If the object is not owned, return None.

        o If the owner's user database doesn't exist, return Nobody.

        o If the owner ID does not exist in the user database, return Nobody.
        """
        owner = self.getOwnerTuple()

        if owner is None or owner is UnownableOwner:
            return None

        udb_path, oid = owner

        root = self.getPhysicalRoot()
        udb = root.unrestrictedTraverse(udb_path, None)

        if udb is None:
            return SU.nobody

        user = udb.getUserById(oid, None)

        if user is None:
            return SU.nobody

        return user.__of__(udb)

    @security.private
    def changeOwnership(self, user, recursive=0):
        """Change the ownership to the given user.

        If 'recursive' is true then also take ownership of all sub-objects,
        otherwise sub-objects retain their ownership information.
        """
        new = ownerInfo(user)
        if new is None:
            return  # Special user!
        old = self.getOwnerTuple()

        if not recursive:
            if old == new or old is UnownableOwner:
                return

        if recursive:
            children = getattr(aq_base(self), 'objectValues', lambda: ())()
            for child in children:
                child.changeOwnership(user, 1)

        if old is not UnownableOwner:
            self._owner = new

    def userCanTakeOwnership(self):
        security = getSecurityManager()
        user = security.getUser()
        info = ownerInfo(user)
        if info is None:
            return 0
        owner = self.getOwnerTuple()
        if owner == info:
            return 0
        return security.checkPermission('Take ownership', self)

    def _deleteOwnershipAfterAdd(self):
        # Only delete _owner if it is an instance attribute.
        if self.__dict__.get('_owner', _mark) is not _mark:
            del self._owner

        for object in self.objectValues():
            try:
                s = object._p_changed
            except BaseException:
                s = 0
            try:
                object._deleteOwnershipAfterAdd()
            except BaseException:
                pass
            if s is None:
                object._p_deactivate()

    def manage_fixupOwnershipAfterAdd(self):
        # Sigh, get the parent's _owner
        parent = getattr(self, '__parent__', None)
        if parent is not None:
            _owner = aq_get(parent, '_owner', None, 1)
        else:
            _owner = None

        if _owner is None and \
           (getattr(self, '__parent__', None) is None or  # NOQA: W504
                not hasattr(self, 'getPhysicalRoot')):
            # This is a special case. An object is
            # being added to an object that hasn't
            # been added to the object hierarchy yet.
            # We can delay fixing up the ownership until the
            # object is actually added.
            return None

        if _owner is UnownableOwner:
            # We want to acquire Unownable ownership!
            return self._deleteOwnershipAfterAdd()
        else:
            # Otherwise change the ownership
            user = getSecurityManager().getUser()
            if (SU.emergency_user and aq_base(user) is SU.emergency_user):
                __creatable_by_emergency_user__ = getattr(
                    self, '__creatable_by_emergency_user__', None)
                if __creatable_by_emergency_user__ is None or \
                        not __creatable_by_emergency_user__():
                    raise EmergencyUserCannotOwn(
                        "Objects cannot be owned by the emergency user")
            self.changeOwnership(user)

        # Force all subs to acquire ownership!
        for object in self.objectValues():
            try:
                s = object._p_changed
            except BaseException:
                s = 0
            try:
                object._deleteOwnershipAfterAdd()
            except BaseException:
                pass
            if s is None:
                object._p_deactivate()


InitializeClass(Owned)


class EmergencyUserCannotOwn(Exception):
    "The emergency user cannot own anything"


class EditUnowned(Exception):
    "Can't edit unowned executables"


def absattr(attr):
    if callable(attr):
        return attr()
    return attr


def ownerInfo(user, getattr=getattr):
    if user is None:
        return None
    uid = user.getId()
    if uid is None:
        return uid
    db = aq_parent(aq_inner(user))
    if db is None:
        return None
    path = [absattr(db.id)]
    root = db.getPhysicalRoot()
    while True:
        db = getattr(db, 'aq_inner', None)
        if db is None:
            break
        db = aq_parent(db)
        if db is root:
            break
        id = db.id
        if not isinstance(id, str):
            try:
                id = id()
            except BaseException:
                id = str(id)
        path.append(id)

    path.reverse()

    return path, uid



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Access control exceptions
"""

import zExceptions


class Unauthorized(zExceptions.Unauthorized):

    def getValueName(self):
        v = self.value
        n = getattr(v, 'getId', v)
        if n is v:
            n = getattr(v, 'id', v)
        if n is v:
            n = getattr(v, '__name__', v)
        if n is not v:
            if callable(n):
                try:
                    n = n()
                except TypeError:
                    pass
            return n

        c = getattr(v, '__class__', type(v))
        c = getattr(c, '__name__', 'object')
        return "a particular %s" % c



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test Zope Guards
"""

import doctest
import gc
import operator
import os
import sys
import unittest

import six

from AccessControl.ZopeGuards import guarded_all
from AccessControl.ZopeGuards import guarded_any
from AccessControl.ZopeGuards import guarded_getattr


try:
    __file__
except NameError:
    __file__ = os.path.abspath(sys.argv[1])
_FILEPATH = os.path.abspath(__file__)
_HERE = os.path.dirname(_FILEPATH)


class SecurityManager:

    def __init__(self, reject=0):
        self.calls = []
        self.reject = reject

    def validate(self, *args):
        from AccessControl import Unauthorized
        self.calls.append(('validate', args))
        if self.reject:
            raise Unauthorized
        return 1

    def validateValue(self, *args):
        from AccessControl import Unauthorized
        self.calls.append(('validateValue', args))
        if self.reject:
            raise Unauthorized
        return 1

    def checkPermission(self, *args):
        self.calls.append(('checkPermission', args))
        return not self.reject


class GuardTestCase(unittest.TestCase):

    def setSecurityManager(self, manager):
        from AccessControl.SecurityManagement import _managers
        from AccessControl.SecurityManagement import get_ident
        key = get_ident()
        old = _managers.get(key)
        if manager is None:
            del _managers[key]
        else:
            _managers[key] = manager
        return old


class Method:

    def __init__(self, *args):
        self.args = args


class TestGuardedGetattr(GuardTestCase):

    def setUp(self):
        self.__sm = SecurityManager()
        self.__old = self.setSecurityManager(self.__sm)
        gc.disable()

    def tearDown(self):
        self.setSecurityManager(self.__old)
        gc.enable()

    def test_miss(self):
        from AccessControl.ZopeGuards import guarded_getattr
        obj, name = object(), 'nonesuch'
        self.assertRaises(AttributeError, guarded_getattr, obj, name)
        self.assertEqual(len(self.__sm.calls), 0)

    def test_unhashable_key(self):
        from AccessControl.ZopeGuards import guarded_getattr
        obj, name = object(), {}
        self.assertRaises(TypeError, guarded_getattr, obj, name)
        self.assertEqual(len(self.__sm.calls), 0)

    def test_unauthorized(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_getattr
        obj, name = Method(), 'args'
        value = getattr(obj, name)
        rc = sys.getrefcount(value)
        self.__sm.reject = True
        self.assertRaises(Unauthorized, guarded_getattr, obj, name)
        self.assertEqual(len(self.__sm.calls), 1)
        del self.__sm.calls[:]
        self.assertEqual(rc, sys.getrefcount(value))

    def test_calls_validate_for_unknown_type(self):
        from AccessControl.ZopeGuards import guarded_getattr
        guarded_getattr(self, 'test_calls_validate_for_unknown_type')
        self.assertEqual(len(self.__sm.calls), 1)

    def test_attr_handler_table(self):
        from AccessControl import Unauthorized
        from AccessControl.SimpleObjectPolicies import ContainerAssertions
        from AccessControl.ZopeGuards import guarded_getattr
        d = {}
        _dict = type(d)
        old = ContainerAssertions.get(_dict)

        mytable = {
            'keys': 1,
            'values': Method,
        }
        ContainerAssertions[_dict] = mytable
        try:
            guarded_getattr(d, 'keys')
            self.assertEqual(len(self.__sm.calls), 0)
            values = guarded_getattr(d, 'values')
            self.assertEqual(values.__class__, Method)
            self.assertEqual(values.args, (d, 'values'))
            self.assertRaises(Unauthorized, guarded_getattr, d, 'items')
        finally:
            ContainerAssertions[_dict] = old

    def test_simple_object_policies(self):
        '''
        Test that we are able to access attributes of simple types (here:
        unicode and bytes)
        '''
        from AccessControl.ZopeGuards import guarded_getattr
        orig_value = self.__sm.reject
        self.__sm.reject = True
        try:
            items = [b'a ', u'a ']
            for item in items:
                self.assertEqual(guarded_getattr(item, 'strip')(),
                                 item.strip())
        finally:
            self.__sm.reject = orig_value
        self.assertEqual(len(self.__sm.calls), 0)


class TestGuardedHasattr(GuardTestCase):

    def setUp(self):
        self.__sm = SecurityManager()
        self.__old = self.setSecurityManager(self.__sm)
        gc.disable()

    def tearDown(self):
        self.setSecurityManager(self.__old)
        gc.enable()

    def test_miss(self):
        from AccessControl.ZopeGuards import guarded_hasattr
        obj, name = object(), 'nonesuch'
        self.assertFalse(guarded_hasattr(obj, name))
        self.assertEqual(len(self.__sm.calls), 0)
        del self.__sm.calls[:]

    def test_unhashable_key(self):
        from AccessControl.ZopeGuards import guarded_hasattr
        obj, name = object(), {}
        self.assertFalse(guarded_hasattr(obj, name))
        self.assertEqual(len(self.__sm.calls), 0)

    def test_unauthorized(self):
        from AccessControl.ZopeGuards import guarded_hasattr
        obj, name = Method(), 'args'
        value = getattr(obj, name)
        rc = sys.getrefcount(value)
        self.__sm.reject = True
        self.assertFalse(guarded_hasattr(obj, name))
        self.assertEqual(len(self.__sm.calls), 1)
        del self.__sm.calls[:]
        self.assertEqual(rc, sys.getrefcount(value))

    def test_hit(self):
        from AccessControl.ZopeGuards import guarded_hasattr
        obj = Method(2411)
        name = 'args'
        value = getattr(obj, name)
        rc = sys.getrefcount(value)
        self.assertTrue(guarded_hasattr(obj, name))
        self.assertEqual(len(self.__sm.calls), 1)
        del self.__sm.calls[:]
        self.assertEqual(rc, sys.getrefcount(value))


class TestDictGuards(GuardTestCase):

    def test_get_simple(self):
        from AccessControl.ZopeGuards import get_dict_get
        get = get_dict_get({'foo': 'bar'}, 'get')
        self.assertEqual(get('foo'), 'bar')

    def test_get_default(self):
        from AccessControl.ZopeGuards import get_dict_get
        get = get_dict_get({'foo': 'bar'}, 'get')
        self.assertTrue(get('baz') is None)
        self.assertEqual(get('baz', 'splat'), 'splat')

    def test_get_validates(self):
        from AccessControl.ZopeGuards import get_dict_get
        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        get = get_dict_get({'foo': GuardTestCase}, 'get')
        try:
            get('foo')
        finally:
            self.setSecurityManager(old)
        self.assertTrue(sm.calls)

    def test_pop_simple(self):
        from AccessControl.ZopeGuards import get_dict_pop
        pop = get_dict_pop({'foo': 'bar'}, 'pop')
        self.assertEqual(pop('foo'), 'bar')

    def test_pop_raises(self):
        from AccessControl.ZopeGuards import get_dict_pop
        pop = get_dict_pop({'foo': 'bar'}, 'pop')
        self.assertRaises(KeyError, pop, 'baz')

    def test_pop_default(self):
        from AccessControl.ZopeGuards import get_dict_pop
        pop = get_dict_pop({'foo': 'bar'}, 'pop')
        self.assertEqual(pop('baz', 'splat'), 'splat')

    def test_pop_validates(self):
        from AccessControl.ZopeGuards import get_dict_get
        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        pop = get_dict_get({'foo': GuardTestCase}, 'pop')
        try:
            pop('foo')
        finally:
            self.setSecurityManager(old)
        self.assertTrue(sm.calls)

    @unittest.skipIf(six.PY3, "iter... is Python 2 only")
    def test_iterkeys_simple(self):
        from AccessControl.ZopeGuards import get_iter
        d = {
            'foo': 1,
            'bar': 2,
            'baz': 3,
        }
        iterkeys = get_iter(d, 'iterkeys')
        keys = d.keys()
        keys.sort()
        ikeys = sorted(iterkeys())
        self.assertEqual(keys, ikeys)

    @unittest.skipIf(six.PY3, "iter... is Python 2 only")
    def test_iterkeys_empty(self):
        from AccessControl.ZopeGuards import get_iter
        iterkeys = get_iter({}, 'iterkeys')
        self.assertEqual(list(iterkeys()), [])

    @unittest.skipIf(six.PY2, "keys() is only in Python 3 a generator")
    def test_keys_empty(self):
        from AccessControl.ZopeGuards import get_iter
        keys = get_iter({}, 'keys')
        self.assertEqual(list(keys()), [])

    @unittest.skipIf(six.PY3, "iter... is Python 2 only")
    def test_iterkeys_validates(self):
        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        iterkeys = guarded_getattr({GuardTestCase: 1}, 'iterkeys')
        try:
            next(iterkeys())
        finally:
            self.setSecurityManager(old)
        self.assertTrue(sm.calls)

    @unittest.skipIf(six.PY2, "keys() is only in Python 3 a generator")
    def test_keys_validates(self):
        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        keys = guarded_getattr({GuardTestCase: 1}, 'keys')
        try:
            next(keys())
        finally:
            self.setSecurityManager(old)
        self.assertTrue(sm.calls)

    @unittest.skipIf(six.PY3, "iter... is Python 2 only")
    def test_itervalues_simple(self):
        from AccessControl.ZopeGuards import get_iter
        d = {
            'foo': 1,
            'bar': 2,
            'baz': 3,
        }
        itervalues = get_iter(d, 'itervalues')
        values = d.values()
        values.sort()
        ivalues = sorted(itervalues())
        self.assertEqual(values, ivalues)

    @unittest.skipIf(six.PY3, "iter... is Python 2 only")
    def test_itervalues_empty(self):
        from AccessControl.ZopeGuards import get_iter
        itervalues = get_iter({}, 'itervalues')
        self.assertEqual(list(itervalues()), [])

    @unittest.skipIf(six.PY2, "values() is only in Python 3 a generator")
    def test_values_empty(self):
        from AccessControl.ZopeGuards import get_iter
        values = get_iter({}, 'values')
        self.assertEqual(list(values()), [])

    @unittest.skipIf(six.PY3, "iter... is Python 2 only")
    def test_itervalues_validates(self):
        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        itervalues = guarded_getattr({GuardTestCase: 1}, 'itervalues')
        try:
            next(itervalues())
        finally:
            self.setSecurityManager(old)
        self.assertTrue(sm.calls)

    @unittest.skipIf(six.PY2, "values() is only in Python 3 a generator")
    def test_values_validates(self):
        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        values = guarded_getattr({GuardTestCase: 1}, 'values')
        try:
            next(values())
        finally:
            self.setSecurityManager(old)
        self.assertTrue(sm.calls)

    def test_kvi_iteration(self):
        from AccessControl.ZopeGuards import SafeIter
        d = dict(a=1, b=2)
        for attr in ("keys", "values", "items"):
            v = getattr(d, attr)()
            si = SafeIter(v)
            self.assertEqual(next(si), next(iter(v)))


class TestListGuards(GuardTestCase):

    def test_pop_simple(self):
        from AccessControl.ZopeGuards import get_list_pop
        pop = get_list_pop(['foo', 'bar', 'baz'], 'pop')
        self.assertEqual(pop(), 'baz')
        self.assertEqual(pop(0), 'foo')

    def test_pop_raises(self):
        from AccessControl.ZopeGuards import get_list_pop
        pop = get_list_pop([], 'pop')
        self.assertRaises(IndexError, pop)

    def test_pop_validates(self):
        from AccessControl.ZopeGuards import get_list_pop
        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        pop = get_list_pop([GuardTestCase], 'pop')
        try:
            pop()
        finally:
            self.setSecurityManager(old)
        self.assertTrue(sm.calls)


class TestBuiltinFunctionGuards(GuardTestCase):

    def test_zip_fails(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_zip
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        self.assertRaises(Unauthorized, guarded_zip, [1, 2, 3], [3, 2, 1])
        self.assertRaises(Unauthorized, guarded_zip, [1, 2, 3], [1])
        self.setSecurityManager(old)

    def test_map_fails(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_map
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        self.assertRaises(Unauthorized, guarded_map, str,
                          [1, 2, 3])
        self.assertRaises(Unauthorized, guarded_map, lambda x, y: x + y,
                          [1, 2, 3], [3, 2, 1])
        self.setSecurityManager(old)

    def test_all_fails(self):
        from AccessControl import Unauthorized
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        self.assertRaises(Unauthorized, guarded_all, [True, True, False])
        self.setSecurityManager(old)

    def test_any_fails(self):
        from AccessControl import Unauthorized
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        self.assertRaises(Unauthorized, guarded_any, [True, True, False])
        self.setSecurityManager(old)

    def test_min_fails(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_min
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        self.assertRaises(Unauthorized, guarded_min, [1, 2, 3])
        self.assertRaises(Unauthorized, guarded_min, 1, 2, 3)

        class MyDict(dict):  # guard() skips 'dict' values
            pass

        self.assertRaises(Unauthorized, guarded_min,
                          MyDict(x=1), MyDict(x=2),
                          key=operator.itemgetter('x'))
        self.setSecurityManager(old)

    def test_max_fails(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_max
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        self.assertRaises(Unauthorized, guarded_max, [1, 2, 3])
        self.assertRaises(Unauthorized, guarded_max, 1, 2, 3)

        class MyDict(dict):  # guard() skips 'dict' values
            pass

        self.assertRaises(Unauthorized, guarded_max,
                          MyDict(x=1), MyDict(x=2),
                          key=operator.itemgetter('x'))
        self.setSecurityManager(old)

    def test_enumerate_fails(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_enumerate
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        enum = guarded_enumerate([1, 2, 3])
        self.assertRaises(Unauthorized, enum.next)
        self.setSecurityManager(old)

    def test_sum_fails(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_sum
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        self.assertRaises(Unauthorized, guarded_sum, [1, 2, 3])
        self.setSecurityManager(old)

    def test_zip_succeeds(self):
        from AccessControl.ZopeGuards import guarded_zip
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        self.assertEqual(guarded_zip([1, 2, 3],
                                     [3, 2, 1],
                                     ),
                         [(1, 3), (2, 2), (3, 1)])
        self.assertEqual(guarded_zip([1, 2, 3], [1]), [(1, 1)])
        self.setSecurityManager(old)

    def test_map_succeeds(self):
        from AccessControl.ZopeGuards import guarded_map
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        self.assertEqual(guarded_map(str, [1, 2, 3]), ['1', '2', '3'])
        self.assertEqual(guarded_map(lambda x, y: x + y, [1, 2, 3], [3, 2, 1]),
                         [4, 4, 4])
        self.setSecurityManager(old)

    def test_all_succeeds(self):
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        self.assertEqual(guarded_all([True, True, False]), False)
        self.setSecurityManager(old)

    def test_any_succeeds(self):
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        self.assertEqual(guarded_any([True, True, False]), True)
        self.setSecurityManager(old)

    def test_min_succeeds(self):
        from AccessControl.ZopeGuards import guarded_min
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        self.assertEqual(guarded_min([1, 2, 3]), 1)
        self.assertEqual(guarded_min(1, 2, 3), 1)

        class MyDict(dict):  # guard() skips 'dict' values
            pass

        self.assertEqual(guarded_min(MyDict(x=1), MyDict(x=2),
                                     key=operator.itemgetter('x')),
                         {'x': 1})
        self.setSecurityManager(old)

    def test_max_succeeds(self):
        from AccessControl.ZopeGuards import guarded_max
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        self.assertEqual(guarded_max([1, 2, 3]), 3)
        self.assertEqual(guarded_max(1, 2, 3), 3)

        class MyDict(dict):  # guard() skips 'dict' values
            pass
        self.assertEqual(guarded_max(MyDict(x=1), MyDict(x=2),
                                     key=operator.itemgetter('x')),
                         {'x': 2})
        self.setSecurityManager(old)

    def test_enumerate_succeeds(self):
        from AccessControl.ZopeGuards import guarded_enumerate
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        enum = guarded_enumerate([1, 2, 3])
        self.assertEqual(next(enum), (0, 1))
        self.assertEqual(next(enum), (1, 2))
        self.assertEqual(next(enum), (2, 3))
        self.assertRaises(StopIteration, enum.next)
        self.setSecurityManager(old)

    def test_sum_succeeds(self):
        from AccessControl.ZopeGuards import guarded_sum
        sm = SecurityManager()  # accepts
        old = self.setSecurityManager(sm)
        self.assertEqual(guarded_sum([1, 2, 3]), 6)
        self.assertEqual(guarded_sum([1, 2, 3], start=36), 42)
        self.setSecurityManager(old)

    def test_apply(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import safe_builtins
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        gapply = safe_builtins['apply']

        def f(a=1, b=2):
            return a + b

        # This one actually succeeds, because apply isn't given anything
        # to unpack.
        self.assertEqual(gapply(f), 3)
        # Likewise, because the things passed are empty.
        self.assertEqual(gapply(f, (), {}), 3)

        self.assertRaises(Unauthorized, gapply, f, [1])
        self.assertRaises(Unauthorized, gapply, f, (), {'a': 2})
        self.assertRaises(Unauthorized, gapply, f, [1], {'a': 2})

        sm = SecurityManager()  # accepts
        self.setSecurityManager(sm)
        self.assertEqual(gapply(f), 3)
        self.assertEqual(gapply(f, (), {}), 3)
        self.assertEqual(gapply(f, [0]), 2)
        self.assertEqual(gapply(f, [], {'b': 18}), 19)
        self.assertEqual(gapply(f, [10], {'b': 1}), 11)

        self.setSecurityManager(old)


class TestGuardedDictListTypes(unittest.TestCase):

    def testDictCreation(self):
        from AccessControl.ZopeGuards import safe_builtins
        d = safe_builtins['dict']
        self.assertEqual(d(), {})
        self.assertEqual(d({1: 2}), {1: 2})
        self.assertEqual(d(((1, 2),)), {1: 2})
        self.assertEqual(d(foo=1), {"foo": 1})
        self.assertEqual(d.fromkeys((1, 2, 3)), {1: None, 2: None, 3: None})
        self.assertEqual(d.fromkeys((1, 2, 3), 'f'), {1: 'f', 2: 'f', 3: 'f'})

    def testListCreation(self):
        from AccessControl.ZopeGuards import safe_builtins
        safe_l = safe_builtins['list']
        self.assertEqual(safe_l(), [])
        self.assertEqual(safe_l([1, 2, 3]), [1, 2, 3])
        x = [3, 2, 1]
        self.assertEqual(safe_l(x), [3, 2, 1])
        self.assertEqual(sorted(safe_l(x)), [1, 2, 3])


class TestRestrictedPythonApply(GuardTestCase):

    def test_apply(self):
        from AccessControl import Unauthorized
        from AccessControl.ZopeGuards import guarded_apply
        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        gapply = guarded_apply

        def f(a=1, b=2):
            return a + b

        # This one actually succeeds, because apply isn't given anything
        # to unpack.
        self.assertEqual(gapply(*(f,)), 3)
        # Likewise, because the things passed are empty.
        self.assertEqual(gapply(*(f,), **{}), 3)

        self.assertRaises(Unauthorized, gapply, *(f, 1))
        self.assertRaises(Unauthorized, gapply, *(f,), **{'a': 2})
        self.assertRaises(Unauthorized, gapply, *(f, 1), **{'a': 2})

        sm = SecurityManager()  # accepts
        self.setSecurityManager(sm)
        self.assertEqual(gapply(*(f,)), 3)
        self.assertEqual(gapply(*(f,), **{}), 3)
        self.assertEqual(gapply(*(f, 0)), 2)
        self.assertEqual(gapply(*(f,), **{'b': 18}), 19)
        self.assertEqual(gapply(*(f, 10), **{'b': 1}), 11)

        self.setSecurityManager(old)


# Map function name to the # of times it's been called.
wrapper_count = {}


class FuncWrapper:
    def __init__(self, funcname, func):
        self.funcname = funcname
        wrapper_count[funcname] = 0
        self.func = func

    def __call__(self, *args, **kws):
        wrapper_count[self.funcname] += 1
        return self.func(*args, **kws)

    def __repr__(self):
        return "<FuncWrapper around %r>" % self.func


# Given the high wall between AccessControl and RestrictedPython, I suppose
# the next one could be called an integration test.
# But we're simply trying to run RestrictedPython with the *intended*
# implementations of the special wrappers here, so no apologies.
_ProtectedBase = None


class TestActualPython(GuardTestCase):

    _old_mgr = _old_policy = _marker = []

    def setUp(self):
        self._wrapped_dicts = []

    def tearDown(self):
        self._restorePolicyAndManager()
        for munged, orig in self._wrapped_dicts:
            munged.update(orig)
        del self._wrapped_dicts

    def _initPolicyAndManager(self, manager=None):
        from AccessControl.SecurityManagement import _managers
        from AccessControl.SecurityManagement import get_ident
        from AccessControl.SecurityManagement import newSecurityManager
        from AccessControl.SecurityManager import setSecurityPolicy
        from AccessControl.ZopeSecurityPolicy import ZopeSecurityPolicy

        class UnderprivilegedUser:
            """ Anonymous USer for unit testing purposes.
            """

            def getId(self):
                return 'Underprivileged User'

            getUserName = getId

            def allowed(self, object, object_roles=None):
                return 0

            def getRoles(self):
                return ()

        self._policy = ZopeSecurityPolicy()
        self._old_policy = setSecurityPolicy(self._policy)

        if manager is None:
            thread_id = get_ident()
            self._old_mgr = manager = _managers.get(thread_id, self._marker)
            newSecurityManager(None, UnderprivilegedUser())
        else:
            self._old_mgr = self.setSecurityManager(manager)

    def _restorePolicyAndManager(self):
        from AccessControl.SecurityManagement import noSecurityManager
        from AccessControl.SecurityManager import setSecurityPolicy

        if self._old_mgr is not self._marker:
            self.setSecurityManager(self._old_mgr)
        else:
            noSecurityManager()

        if self._old_policy is not self._marker:
            setSecurityPolicy(self._old_policy)

    def _getProtectedBaseClass(self):
        from ExtensionClass import Base

        from AccessControl.class_init import InitializeClass
        from AccessControl.SecurityInfo import ClassSecurityInfo

        global _ProtectedBase
        if _ProtectedBase is None:

            class ProtectedBase(Base):
                security = ClassSecurityInfo()

                @security.private
                def private_method(self):
                    return 'private_method called'

            InitializeClass(ProtectedBase)
            _ProtectedBase = ProtectedBase

        return _ProtectedBase

    def testPython(self):
        code, its_globals = self._compile("actual_python.py")

        # Fiddle the global and safe-builtins dicts to count how many times
        # the special functions are called.
        self._wrap_replaced_dict_callables(its_globals)
        self._wrap_replaced_dict_callables(its_globals['__builtins__'])

        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        try:
            exec(code, its_globals)
        finally:
            self.setSecurityManager(old)

        # Use wrapper_count to determine coverage.
        # print wrapper_count  # uncomment to see wrapper names & counts
        untouched = [k for k, v in wrapper_count.items() if v == 0]
        if untouched:
            untouched.sort()
            self.fail("Unexercised wrappers: %r" % untouched)

    def testPythonRealAC(self):
        code, its_globals = self._compile("actual_python.py")
        exec(code, its_globals)

    def test_derived_class_normal(self):
        from AccessControl import Unauthorized

        NORMAL_SCRIPT = """
class Normal(ProtectedBase):
    pass

normal = Normal()
result = normal.private_method()
"""
        code, its_globals = self._compile_str(NORMAL_SCRIPT, 'normal_script')
        its_globals['ProtectedBase'] = self._getProtectedBaseClass()

        self._initPolicyAndManager()

        with self.assertRaises(Unauthorized):
            exec(code, its_globals)
            self.fail("Didn't raise Unauthorized: \n%s" %
                      its_globals['result']())

    def test_derived_class_sneaky_en_suite(self):
        #  Disallow declaration of security-affecting names in classes
        #  defined in restricted code (compile-time check).

        SNEAKY_SCRIPT = """
class Sneaky(ProtectedBase):
    private_method__roles__ = None


sneaky = Sneaky()
result = sneaky.private_method()
"""
        with self.assertRaises(SyntaxError):
            self._compile_str(SNEAKY_SCRIPT, 'sneaky_script')
            self.fail("Didn't raise SyntaxError!")

    def test_derived_sneaky_post_facto(self):
        #  Assignment to a class outside its suite fails at
        #  compile time with a SyntaxError.

        SNEAKY_SCRIPT = """
class Sneaky(ProtectedBase):
    pass

Sneaky.private_method__roles__ = None

sneaky = Sneaky()
result = sneaky.private_method()
"""
        with self.assertRaises(SyntaxError):
            self._compile_str(SNEAKY_SCRIPT, 'sneaky_script')
            self.fail("Didn't raise SyntaxError!")

    def test_derived_sneaky_instance(self):
        #  Assignment of security-sensitive names to an instance
        #  fails at compile time with a SyntaxError.

        SNEAKY_SCRIPT = """
class Sneaky(ProtectedBase):
    pass

sneaky = Sneaky()
sneaky.private_method__roles__ = None
result = sneaky.private_method()
"""
        with self.assertRaises(SyntaxError):
            self._compile_str(SNEAKY_SCRIPT, 'sneaky_script')
            self.fail("Didn't raise SyntaxError!")

    def test_dict_access(self):
        SIMPLE_DICT_ACCESS_SCRIPT = """
def foo(text):
    return text

kw = {'text':'baz'}
print(foo(**kw))

kw = {'text':True}
print(foo(**kw))
printed  # Prevent a warning of RestrictedPython that itis not used.
"""
        code, its_globals = self._compile_str(SIMPLE_DICT_ACCESS_SCRIPT, 'x')

        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        try:
            exec(code, its_globals)
        finally:
            self.setSecurityManager(old)

        self.assertEqual(its_globals['_print'](),
                         'baz\nTrue\n')

    def test_guarded_next__1(self):
        """There is a `safe_builtin` named `next`."""
        SCRIPT = "result = next(iterator)"

        code, its_globals = self._compile_str(SCRIPT, 'ignored')
        its_globals['iterator'] = iter([2411, 123])

        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        try:
            exec(code, its_globals)
        finally:
            self.setSecurityManager(old)

        self.assertEqual(its_globals['result'], 2411)

    def test_guarded_next_default(self):
        """There is a `safe_builtin` named `next` with `default`."""
        SCRIPT = "result = next(iterator, 'default')"

        code, its_globals = self._compile_str(SCRIPT, 'ignored')
        its_globals['iterator'] = iter([])

        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        try:
            exec(code, its_globals)
        finally:
            self.setSecurityManager(old)
        self.assertEqual(its_globals['result'], "default")

    def test_guarded_next_StopIteration(self):
        """There is a `safe_builtin` named `next`, raising StopIteration
        when iterator is exhausted."""
        SCRIPT = "result = next(iterator)"

        code, its_globals = self._compile_str(SCRIPT, 'ignored')
        its_globals['iterator'] = iter([])

        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        with self.assertRaises(StopIteration):
            try:
                exec(code, its_globals)
            finally:
                self.setSecurityManager(old)

    def test_guarded_next__2(self):
        """It guards the access during iteration."""
        from AccessControl import Unauthorized

        SCRIPT = "next(iterator)"

        code, its_globals = self._compile_str(SCRIPT, 'ignored')
        its_globals['iterator'] = iter([2411, 123])

        sm = SecurityManager(reject=True)
        old = self.setSecurityManager(sm)
        with self.assertRaises(Unauthorized):
            try:
                exec(code, its_globals)
            finally:
                self.setSecurityManager(old)
        self.assertEqual([('validate', (2411, 2411, None, 2411))], sm.calls)

    def test_guarded_next__3(self):
        """It does not double check if using a `SafeIter`."""
        from AccessControl.ZopeGuards import SafeIter

        SCRIPT = "next(iter([2411, 123]))"

        code, its_globals = self._compile_str(SCRIPT, 'ignored')
        its_globals['iterator'] = SafeIter([2411, 123])

        sm = SecurityManager()
        old = self.setSecurityManager(sm)
        try:
            exec(code, its_globals)
        finally:
            self.setSecurityManager(old)
        self.assertEqual(
            [('validate', ([2411, 123], [2411, 123], None, 2411))], sm.calls)

    def _compile_str(self, text, name):
        from RestrictedPython import compile_restricted

        from AccessControl.ZopeGuards import get_safe_globals
        from AccessControl.ZopeGuards import guarded_getattr

        code = compile_restricted(text, name, 'exec')

        g = get_safe_globals()
        g['_getattr_'] = guarded_getattr
        g['__debug__'] = 1  # so assert statements are active
        g['__name__'] = __name__  # so classes can be defined in the script
        return code, g

    # Compile code in fname, as restricted Python. Return the
    # compiled code, and a safe globals dict for running it in.
    # fname is the string name of a Python file; it must be found
    # in the same directory as this file.
    def _compile(self, fname):
        fn = os.path.join(_HERE, fname)
        with open(fn) as f:
            text = f.read()
        return self._compile_str(text, fn)

    # d is a dict, the globals for execution or our safe builtins.
    # The callable values which aren't the same as the corresponding
    # entries in builtins are wrapped in a FuncWrapper, so we can
    # tell whether they're executed.
    def _wrap_replaced_dict_callables(self, d):
        from RestrictedPython.Guards import builtins
        orig = d.copy()
        self._wrapped_dicts.append((d, orig))
        for k, v in d.items():
            if callable(v) and v is not getattr(builtins, k, None):
                d[k] = FuncWrapper(k, v)


def test_inplacevar():
    """
Verify the correct behavior of protected_inplacevar.

    >>> from AccessControl.ZopeGuards import protected_inplacevar

Basic operations on objects without inplace slots work as expected:

    >>> protected_inplacevar('+=', 1, 2)
    3
    >>> protected_inplacevar('-=', 5, 2)
    3
    >>> protected_inplacevar('*=', 5, 2)
    10
    >>> protected_inplacevar('/=', 6, 2.0)
    3.0
    >>> protected_inplacevar('//=', 6, 2)
    3
    >>> protected_inplacevar('%=', 5, 2)
    1
    >>> protected_inplacevar('**=', 5, 2)
    25
    >>> protected_inplacevar('<<=', 5, 2)
    20
    >>> protected_inplacevar('>>=', 5, 2)
    1
    >>> protected_inplacevar('&=', 5, 2)
    0
    >>> protected_inplacevar('^=', 7, 2)
    5
    >>> protected_inplacevar('|=', 5, 2)
    7

Inplace operations are allowed on lists:

    >>> protected_inplacevar('+=', [1], [2])
    [1, 2]

    >>> protected_inplacevar('*=', [1], 2)
    [1, 1]

But not on custom objects:

    >>> class C:
    ...     def __iadd__(self, other):
    ...         return 42
    >>> protected_inplacevar('+=', C(), 2)    # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    TypeError: Augmented assignment to C objects is not allowed in
    untrusted code
"""


def test_inplacevar_for_py24():
    """
protected_inplacevar allows in-place ops on sets:

    >>> from AccessControl.ZopeGuards import protected_inplacevar
    >>> s = set((1,2,3,4))
    >>> sorted(protected_inplacevar('-=', s, set((1, 3))))
    [2, 4]
    >>> sorted(s)
    [2, 4]

    >>> sorted(protected_inplacevar('|=', s, set((1, 3, 9))))
    [1, 2, 3, 4, 9]
    >>> sorted(s)
    [1, 2, 3, 4, 9]

    >>> sorted(protected_inplacevar('&=', s, set((1, 2, 3, 9))))
    [1, 2, 3, 9]
    >>> sorted(s)
    [1, 2, 3, 9]

    >>> sorted(protected_inplacevar('^=', s, set((1, 3, 7, 8))))
    [2, 7, 8, 9]
    >>> sorted(s)
    [2, 7, 8, 9]
"""


def test_suite():
    suite = unittest.TestSuite([
        doctest.DocTestSuite(),
    ])
    for cls in (TestGuardedGetattr,
                TestGuardedHasattr,
                TestDictGuards,
                TestBuiltinFunctionGuards,
                TestListGuards,
                TestGuardedDictListTypes,
                TestRestrictedPythonApply,
                TestActualPython,
                ):
        suite.addTest(unittest.makeSuite(cls))
    return suite



import unittest


def _makeRootAndUser():
    from Acquisition import Explicit
    from Acquisition import Implicit

    from AccessControl.rolemanager import RoleManager

    class DummyContext(Implicit, RoleManager):
        __roles__ = ('Manager',)

    class DummyUser(Explicit):
        def getRoles(self):
            return ('Manager',)

        def getRolesInContext(self, context):
            return ('Manager',)

        def has_permission(self, permission, context):
            return True

    class DummyAclUsers(Explicit):
        def getUser(self, user_id):
            user = DummyUser()
            return user.__of__(self)

        def absolute_url(self, relative=0):
            return 'acl_users'

    class DummyRoot(Explicit):
        acl_users = DummyAclUsers()

    root = DummyRoot()
    root.acl_users = DummyAclUsers()
    root.context1 = DummyContext()
    root.context2 = DummyContext()
    user = DummyUser().__of__(root.acl_users)

    return root, user


class TestRoleManager(unittest.TestCase):

    def tearDown(self):
        from AccessControl.SecurityManagement import noSecurityManager
        noSecurityManager()

    def test_interfaces(self):
        from zope.interface.verify import verifyClass

        from AccessControl.interfaces import IRoleManager
        from AccessControl.rolemanager import RoleManager

        verifyClass(IRoleManager, RoleManager)

    def test_manage_getUserRolesAndPermissions(self):
        from AccessControl.ImplPython import verifyAcquisitionContext
        from AccessControl.SecurityManagement import getSecurityManager
        from AccessControl.SecurityManagement import newSecurityManager
        root, user = _makeRootAndUser()
        newSecurityManager(None, user)
        root.context1.manage_getUserRolesAndPermissions('dummy_user')
        user = getSecurityManager().getUser()
        self.assertTrue(verifyAcquisitionContext(user, root.context2, ()))

    def test_has_local_roles(self):
        root, user = _makeRootAndUser()
        self.assertFalse(root.context1.has_local_roles())

    def test_get_local_roles(self):
        root, user = _makeRootAndUser()
        root.context1.__ac_local_roles__ = {'user1': ['Role1']}
        roles = root.context1.get_local_roles()
        self.assertEqual(roles, (
            ('user1', ('Role1',)),
        ))

    def test_manage_addLocalRoles(self):
        root, user = _makeRootAndUser()
        root.context1.manage_addLocalRoles('user1', ['Role1'])
        roles = root.context1.get_local_roles_for_userid('user1')
        self.assertEqual(roles, ('Role1',))

    def test_manage_setLocalRoles(self):
        root, user = _makeRootAndUser()
        root.context1.__ac_local_roles__ = {'user1': ('Role1',)}
        root.context1.manage_setLocalRoles('user1', ['Role2'])
        roles = root.context1.get_local_roles_for_userid('user1')
        self.assertEqual(roles, ('Role2',))

    def test_manage_delLocalRoles(self):
        root, user = _makeRootAndUser()
        root.context1.__ac_local_roles__ = {'user1': ('Role1',)}
        root.context1.manage_delLocalRoles(['user1'])
        roles = root.context1.get_local_roles_for_userid('user1')
        self.assertEqual(roles, ())

    def test_valid_roles(self):
        from AccessControl.rolemanager import RoleManager
        root, user = _makeRootAndUser()

        # default case, __ac_roles__ not overridden
        self.assertEqual(set(root.context1.valid_roles()),
                         set(RoleManager.__ac_roles__))

        # forcing our own roles
        root.context1.__ac_roles__ = ('Role2', 'Role1')
        roles = root.context1.valid_roles()
        self.assertEqual(roles, ('Role1', 'Role2'))

    def test_validate_roles(self):
        from AccessControl.rolemanager import RoleManager
        root, user = _makeRootAndUser()

        # default case, __ac_roles__ not overridden
        self.assertTrue(root.context1.validate_roles(RoleManager.__ac_roles__))
        self.assertFalse(root.context1.validate_roles(('Role1', 'Role2')))

        # forcing our own roles
        root.context1.__ac_roles__ = ('Role2', 'Role1')
        validator = root.context1.validate_roles
        self.assertFalse(validator(RoleManager.__ac_roles__))
        self.assertTrue(validator(('Role1', 'Role2')))

    def test_userdefined_roles(self):
        root, user = _makeRootAndUser()

        # default case, __ac_roles__ not overridden
        self.assertEqual(root.context1.userdefined_roles(), ())

        # forcing our own roles
        root.context1.__ac_roles__ = ('Role2', 'Role1')
        self.assertEqual(
            ('Role2', 'Role1'),
            root.context1.userdefined_roles(),
        )



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Permission Mapping

Sometimes, we need an object's permissions to be remapped to other permissions
when the object is used in special ways.  This is rather hard, since we
need the object's ordinary permissions intact so we can manage it.
"""

from Acquisition import ImplicitAcquisitionWrapper
from ExtensionClass import Base
from zope.interface import implementer

from AccessControl.class_init import InitializeClass
from AccessControl.interfaces import IPermissionMappingSupport
from AccessControl.owner import UnownableOwner
from AccessControl.Permission import getPermissionIdentifier
from AccessControl.requestmethod import requestmethod


try:
    from html import escape
except ImportError:  # PY2
    from cgi import escape


@implementer(IPermissionMappingSupport)
class RoleManager:

    # XXX: No security declarations?

    def manage_getPermissionMapping(self):
        """Return the permission mapping for the object

        This is a list of dictionaries with:

          permission_name -- The name of the native object permission

          class_permission -- The class permission the permission is
             mapped to.
        """
        wrapper = getattr(self, '_permissionMapper', None)
        if wrapper is None:
            wrapper = PM()

        perms = {}
        for p in self.possible_permissions():
            perms[getPermissionIdentifier(p)] = p

        r = []
        a = r.append
        for ac_perms in self.ac_inherited_permissions(1):
            p = perms.get(getPermissionMapping(ac_perms[0], wrapper), '')
            a({'permission_name': ac_perms[0], 'class_permission': p})
        return r

    @requestmethod('POST')
    def manage_setPermissionMapping(self,
                                    permission_names=[],
                                    class_permissions=[],
                                    REQUEST=None):
        """Change the permission mapping
        """
        wrapper = getattr(self, '_permissionMapper', None)
        if wrapper is None:
            wrapper = PM()

        perms = self.possible_permissions()
        for i in range(len(permission_names)):
            name = permission_names[i]
            p = class_permissions[i]
            if p and (p not in perms):
                __traceback_info__ = perms, p, i
                raise ValueError(
                    """Attempted to map a permission to a permission, %s,
                    that is not valid. This should never happen. (Waaa).
                    """ % escape(p))
            setPermissionMapping(name, wrapper, p)

        self._permissionMapper = wrapper

        if REQUEST is not None:
            return self.manage_access(
                REQUEST,
                manage_tabs_message='The permission mapping has been updated')

    def _isBeingUsedAsAMethod(self, REQUEST=None, wannaBe=0):
        try:
            if hasattr(self, 'aq_self'):
                r = self.aq_acquire('_isBeingUsedAsAMethod_')
            else:
                r = self._isBeingUsedAsAMethod_
        except BaseException:
            r = 0

        if REQUEST is not None:
            if not r != (not wannaBe):
                REQUEST.response.notFoundError()

        return r


InitializeClass(RoleManager)


def getPermissionMapping(name, obj, st=type('')):
    obj = getattr(obj, 'aq_base', obj)
    name = getPermissionIdentifier(name)
    r = getattr(obj, name, '')
    if not isinstance(r, st):
        r = ''
    return r


def setPermissionMapping(name, obj, v):
    name = getPermissionIdentifier(name)
    if v:
        setattr(obj, name, getPermissionIdentifier(v))
    elif name in obj.__dict__:
        delattr(obj, name)


class PM(Base):
    _owner = UnownableOwner

    _View_Permission = '_View_Permission'
    _is_wrapperish = 1

    def __getattr__(self, name):
        # We want to make sure that any non-explicitly set methods are
        # private!
        if name.startswith('_') and name.endswith("_Permission"):
            return ''
        raise AttributeError(escape(name))


PermissionMapper = PM


def aqwrap(object, wrapper, parent):
    r = Rewrapper()
    r._ugh = wrapper, object, parent
    return r


class Rewrapper(Base):
    def __of__(self, parent):
        w, m, p = self._ugh
        return m.__of__(ImplicitAcquisitionWrapper(w, parent))

    def __getattr__(self, name):
        w, m, parent = self._ugh
        self = m.__of__(ImplicitAcquisitionWrapper(w, parent))
        return getattr(self, name)

    def __call__(self, *args, **kw):
        w, m, parent = self._ugh
        self = m.__of__(ImplicitAcquisitionWrapper(w, parent))
        return self(*args, **kw)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from __future__ import absolute_import

import AuthEncoding.AuthEncoding
from zope.deferredimport import deprecated


deprecated(
    "The functionality of AccessControl.AuthEncoding has moved to"
    " AuthEncoding.AuthEncoding. Please import from there."
    " This backward compatibility shim will be removed in AccessControl"
    " version 6.",
    MySQLDigestScheme='AuthEncoding.AuthEncoding:MySQLDigestScheme',
    PasswordEncryptionScheme='AuthEncoding.AuthEncoding:PasswordEncryptionScheme',  # noqa
    SHADigestScheme='AuthEncoding.AuthEncoding:SHADigestScheme',
    SSHADigestScheme='AuthEncoding.AuthEncoding:SSHADigestScheme',
    constant_time_compare='AuthEncoding.AuthEncoding:constant_time_compare',
    is_encrypted='AuthEncoding.AuthEncoding:is_encrypted',
    listSchemes='AuthEncoding.AuthEncoding:listSchemes',
    pw_encode='AuthEncoding.AuthEncoding:pw_encode',
    pw_encrypt='AuthEncoding.AuthEncoding:pw_encrypt',
    pw_validate='AuthEncoding.AuthEncoding:pw_validate',
    registerScheme='AuthEncoding.AuthEncoding:registerScheme',
)

# Bogosity on various platforms due to ITAR restrictions.
# We need to import the module anyway to check for the presence of
# restrictions, but do want to silence the ImportError.

if hasattr(AuthEncoding.AuthEncoding, 'CryptDigestScheme'):
    deprecated(
        "The functionality of AccessControl.AuthEncoding has moved to"
        " AuthEncoding.AuthEncoding. Please import from there."
        " This backward compatibility shim will be removed in AccessControl"
        " version 6.",
        CryptDigestScheme='AuthEncoding.AuthEncoding:CryptDigestScheme',
    )



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for AccessControl.User
"""
import unittest


class BasicUserTests(unittest.TestCase):

    def _getTargetClass(self):
        from AccessControl.users import BasicUser
        return BasicUser

    def _makeOne(self, name, password, roles, domains):
        return self._getTargetClass()(name, password, roles, domains)

    def test_interfaces(self):
        from zope.interface.verify import verifyClass

        from AccessControl.interfaces import IUser

        verifyClass(IUser, self._getTargetClass())

    def _makeDerived(self, **kw):
        class Derived(self._getTargetClass()):
            def __init__(self, **kw):
                self.name = 'name'
                self.password = 'password'
                self.roles = ['Manager']
                self.domains = []
                self.__dict__.update(kw)
        return Derived(**kw)

    def test_ctor_is_abstract(self):
        # Subclasses must override __init__, and mustn't call the base version.
        self.assertRaises(NotImplementedError,
                          self._makeOne, 'name', 'password', ['Manager'], [])

    def test_abstract_methods(self):
        # Subclasses must override these methods.
        derived = self._makeDerived()
        self.assertRaises(NotImplementedError, derived.getUserName)
        self.assertRaises(NotImplementedError, derived.getId)
        self.assertRaises(NotImplementedError, derived._getPassword)
        self.assertRaises(NotImplementedError, derived.getRoles)
        self.assertRaises(NotImplementedError, derived.getDomains)

    def test_getRolesInContext_no_aq_no_local_roles(self):
        derived = self._makeDerived()
        derived.getId = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Manager']
        self.assertEqual(derived.getRolesInContext(self), ['Manager'])

    def test_getRolesInContext_no_aq_w_local_roles_as_dict(self):
        class Target(object):
            __ac_local_roles__ = {'user': ['Other']}

        derived = self._makeDerived()
        derived.getId = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Manager']
        self.assertEqual(derived.getRolesInContext(Target()),
                         ['Manager', 'Other'])

    def test_getRolesInContext_no_aq_w_local_roles_as_callable(self):
        class Context(object):
            def __ac_local_roles__(self):
                return {'user': ['Other']}

        derived = self._makeDerived()
        derived.getId = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Manager']
        self.assertEqual(derived.getRolesInContext(Context()),
                         ['Manager', 'Other'])

    def test_getRolesInContext_w_aq(self):
        class Context(object):
            pass

        derived = self._makeDerived()
        derived.getId = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Manager']
        parent = Context()
        parent.__ac_local_roles__ = {'user': ['Another']}
        target = Context()
        target.__ac_local_roles__ = {'user': ['Other']}
        target.__parent__ = parent
        self.assertEqual(set(derived.getRolesInContext(target)),
                         set(['Manager', 'Other', 'Another']))

    def test_getRolesInContext_w_method(self):
        class Context(object):
            __ac_local_roles__ = {'user': ['Other']}

            def method(self):
                pass

        derived = self._makeDerived()
        derived.getId = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Manager']
        target = Context()
        self.assertEqual(derived.getRolesInContext(target.method),
                         ['Manager', 'Other'])

    def test_authenticate_miss_wo_domains(self):
        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ()
        self.assertFalse(derived.authenticate('notpassword', {}))

    def test_authenticate_hit_wo_domains(self):
        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ()
        self.assertTrue(derived.authenticate('password', {}))

    def test_authenticate_hit_w_domains_any(self):
        class Request(dict):
            def getClientAddr(self):
                return '192.168.1.1'

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('*',)
        request = Request(REMOTE_HOST='example.com')
        self.assertTrue(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_miss_both(self):
        class Request(dict):
            def getClientAddr(self):
                return '192.168.1.1'

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('127.0.0.1',)
        request = Request(REMOTE_HOST='example.com')
        self.assertFalse(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_wo_host_wo_addr(self):
        class Request(dict):
            def getClientAddr(self):
                return ''

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('127.0.0.1',)
        request = Request(REMOTE_HOST='')
        self.assertFalse(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_miss_host_hit_addr(self):
        class Request(dict):
            def getClientAddr(self):
                return '127.0.0.1'

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('127.0.0.1',)
        request = Request(REMOTE_HOST='example.com')
        self.assertTrue(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_miss_host_hit_addr_wo_host(self):
        class Request(dict):
            def getClientAddr(self):
                return '127.0.0.1'

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('127.0.0.1',)
        request = Request(REMOTE_HOST='')
        self.assertTrue(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_miss_host_hit_addr_wo_addr(self):
        class Request(dict):
            def getClientAddr(self):
                return ''

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('127.0.0.1',)
        request = Request(REMOTE_HOST='localhost')
        self.assertTrue(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_miss_addr_hit_host(self):
        class Request(dict):
            def getClientAddr(self):
                return '192.168.1.1'

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('example.com',)
        request = Request(REMOTE_HOST='host.example.com')
        self.assertTrue(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_miss_addr_hit_host_wo_addr(self):
        class Request(dict):
            def getClientAddr(self):
                return ''

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('example.com',)
        request = Request(REMOTE_HOST='host.example.com')
        self.assertTrue(derived.authenticate('password', request))

    def test_authenticate_hit_w_domains_miss_addr_hit_host_wo_host(self):
        class Request(dict):
            def getClientAddr(self):
                return '127.0.0.1'

        derived = self._makeDerived()
        derived._getPassword = lambda *x: 'password'
        derived.getDomains = lambda *x: ('localhost',)
        request = Request(REMOTE_HOST='localhost')
        self.assertTrue(derived.authenticate('password', request))

    def test_allowed__what_not_even_god_should_do(self):
        from AccessControl.PermissionRole import _what_not_even_god_should_do

        derived = self._makeDerived()
        self.assertFalse(derived.allowed(self, _what_not_even_god_should_do))

    def test_allowed_w_roles_is_None(self):
        derived = self._makeDerived()
        self.assertTrue(derived.allowed(self, None))

    def test_allowed_w_Anonymous_in_roles(self):
        derived = self._makeDerived()
        self.assertTrue(derived.allowed(self, ['Anonymous']))

    def test_allowed_w_Authenticated_in_roles_w_Anonymous(self):
        derived = self._makeDerived()
        derived.getUserName = lambda *x: 'Anonymous User'
        derived.getRoles = lambda *x: ['Anonymous']
        self.assertFalse(derived.allowed(self, ['Authenticated']))

    def test_allowed_w_Authenticated_in_roles_not_Anonymous(self):
        derived = self._makeDerived()
        derived.getUserName = lambda *x: 'user'
        self.assertTrue(derived.allowed(self, ['Authenticated']))

    def test_allowed_w_Authenticated_in_roles_not_Anonymous_context_miss(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: []
        self.assertFalse(derived.allowed(self, ['Authenticated']))

    def test_allowed_w_Authenticated_in_roles_not_Anonymous_context_hit(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: []
        self.assertTrue(derived.allowed(folder, ['Authenticated']))

    # TODO: def test_allowed_w_Shared_in_roles XXX rip that out instead

    def test_allowed_w_global_roles_hit_context_miss(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Editor']
        self.assertFalse(derived.allowed(self, ['Editor']))

    def test_allowed_w_global_roles_hit_context_hit(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Editor']
        self.assertTrue(derived.allowed(folder, ['Editor']))

    def test_allowed_w_global_roles_miss_wo_local_roles_wo_aq(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Author']
        self.assertFalse(derived.allowed(self, ['Editor']))

    def test_allowed_w_global_roles_miss_w_local_roles_miss_wo_aq(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        class Folder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        other = Folder()  # no aq
        other.__ac_local_roles__ = {'user': ['Editor']}
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Author']
        self.assertFalse(derived.allowed(other, ['Reviewer']))

    def test_allowed_w_global_roles_miss_w_local_roles_hit_wo_aq(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        class Folder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        other = Folder()  # no aq
        other.__ac_local_roles__ = {'user': ['Editor']}
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Author']
        self.assertFalse(derived.allowed(other, ['Editor']))

    def test_allowed_w_global_roles_miss_w_local_roles_hit_w_aq(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        class Folder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        other = Folder().__of__(root)
        other.__ac_local_roles__ = {'user': ['Editor']}
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: ['Author']
        self.assertTrue(derived.allowed(other, ['Editor']))

    def test_has_role_w_str_wo_context_miss(self):
        derived = self._makeDerived()
        derived.getRoles = lambda *x: ['Author']
        self.assertFalse(derived.has_role('Editor'))

    def test_has_role_w_list_wo_context_hit(self):
        derived = self._makeDerived()
        derived.getRoles = lambda *x: ['Author']
        self.assertTrue(derived.has_role(['Author']))

    def test_has_role_w_list_w_context_miss(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class Folder(Implicit):
            pass

        root = Root()
        root.__ac_local_roles__ = {'user': ['Editor']}
        folder = Folder().__of__(root)
        derived = self._makeDerived()
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: []
        self.assertFalse(derived.has_role(['Author'], folder))

    def test_has_role_w_list_w_context_hit(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class Folder(Implicit):
            pass

        root = Root()
        root.__ac_local_roles__ = {'user': ['Editor']}
        folder = Folder().__of__(root)
        derived = self._makeDerived()
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: []
        self.assertTrue(derived.has_role(['Editor'], folder))

    def test_has_permission_miss(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        derived = self._makeDerived().__of__(folder)
        derived.getUserName = lambda *x: 'user'
        derived.getRoles = lambda *x: []
        self.assertFalse(derived.has_permission('Permission', folder))

    def test_has_permission_hit(self):
        from Acquisition import Implicit

        class Root(Implicit):
            pass

        class UserFolder(Implicit):
            pass

        root = Root()
        folder = UserFolder().__of__(root)
        derived = self._makeDerived().__of__(folder)
        derived.getRoles = lambda *x: ['Manager']
        self.assertTrue(derived.has_permission('Permission', folder))

    def test___len__(self):
        derived = self._makeDerived()
        self.assertEqual(len(derived), 1)

    def test___str__(self):
        derived = self._makeDerived(getUserName=lambda: 'phred')
        self.assertEqual(str(derived), 'phred')

    def test___repr__(self):
        derived = self._makeDerived(getUserName=lambda: 'phred')
        self.assertEqual(repr(derived), "<Derived 'phred'>")


class SimpleUserTests(unittest.TestCase):

    def _getTargetClass(self):
        from AccessControl.users import SimpleUser
        return SimpleUser

    def _makeOne(self, name='admin', password='123', roles=None, domains=None):
        if roles is None:
            roles = ['Manager']
        if domains is None:
            domains = []
        return self._getTargetClass()(name, password, roles, domains)

    def test_interfaces(self):
        from zope.interface.verify import verifyClass

        from AccessControl.interfaces import IUser

        verifyClass(IUser, self._getTargetClass())

    def test_overrides(self):
        simple = self._makeOne()
        self.assertEqual(simple.getUserName(), 'admin')
        self.assertEqual(simple.getId(), 'admin')
        self.assertEqual(simple._getPassword(), '123')
        self.assertEqual(simple.getDomains(), ())

    def test_getRoles_anonymous(self):
        simple = self._makeOne('Anonymous User', roles=())
        self.assertEqual(simple.getRoles(), ())

    def test_getRoles_non_anonymous(self):
        simple = self._makeOne('phred', roles=())
        self.assertEqual(simple.getRoles(), ('Authenticated',))

    def test___repr__(self):
        special = self._makeOne()
        self.assertEqual(repr(special), "<SimpleUser 'admin'>")


class SpecialUserTests(unittest.TestCase):

    def _getTargetClass(self):
        from AccessControl.users import SpecialUser
        return SpecialUser

    def _makeOne(self, name='admin', password='123', roles=None, domains=None):
        if roles is None:
            roles = ['Manager']
        if domains is None:
            domains = []
        return self._getTargetClass()(name, password, roles, domains)

    def test_interfaces(self):
        from zope.interface.verify import verifyClass

        from AccessControl.interfaces import IUser

        verifyClass(IUser, self._getTargetClass())

    def test_overrides(self):
        special = self._makeOne()
        self.assertEqual(special.getUserName(), 'admin')
        self.assertEqual(special.getId(), None)
        self.assertEqual(special._getPassword(), '123')
        self.assertEqual(special.getDomains(), ())

    def test___repr__(self):
        special = self._makeOne()
        self.assertEqual(repr(special), "<SpecialUser 'admin'>")


class UnrestrictedUserTests(unittest.TestCase):

    def _getTargetClass(self):
        from AccessControl.users import UnrestrictedUser
        return UnrestrictedUser

    def _makeOne(self, name='admin', password='123', roles=None, domains=None):
        if roles is None:
            roles = ['Manager']
        if domains is None:
            domains = []
        return self._getTargetClass()(name, password, roles, domains)

    def test_interfaces(self):
        from zope.interface.verify import verifyClass

        from AccessControl.interfaces import IUser

        verifyClass(IUser, self._getTargetClass())

    def test_allowed__what_not_even_god_should_do(self):
        from AccessControl.PermissionRole import _what_not_even_god_should_do
        unrestricted = self._makeOne()
        self.assertFalse(unrestricted.allowed(self,
                                              _what_not_even_god_should_do))

    def test_allowed_empty(self):
        unrestricted = self._makeOne()
        self.assertTrue(unrestricted.allowed(self, ()))

    def test_allowed_other(self):
        unrestricted = self._makeOne()
        self.assertTrue(unrestricted.allowed(self, ('Manager',)))

    def test_has_role_empty_no_object(self):
        unrestricted = self._makeOne()
        self.assertTrue(unrestricted.has_role(()))

    def test_has_role_empty_w_object(self):
        unrestricted = self._makeOne()
        self.assertTrue(unrestricted.has_role((), self))

    def test_has_role_other_no_object(self):
        unrestricted = self._makeOne()
        self.assertTrue(unrestricted.has_role(('Manager',)))

    def test_has_role_other_w_object(self):
        unrestricted = self._makeOne()
        self.assertTrue(unrestricted.has_role(('Manager',), self))

    def test___repr__(self):
        unrestricted = self._makeOne()
        self.assertEqual(repr(unrestricted),
                         "<UnrestrictedUser 'admin'>")


class NullUnrestrictedUserTests(unittest.TestCase):

    def _getTargetClass(self):
        from AccessControl.users import NullUnrestrictedUser
        return NullUnrestrictedUser

    def _makeOne(self):
        return self._getTargetClass()()

    def test_interfaces(self):
        from zope.interface.verify import verifyClass

        from AccessControl.interfaces import IUser

        verifyClass(IUser, self._getTargetClass())

    def test_overrides(self):
        simple = self._makeOne()
        self.assertEqual(simple.getUserName(), (None, None))
        self.assertEqual(simple.getId(), None)
        self.assertEqual(simple._getPassword(), (None, None))
        self.assertEqual(simple.getRoles(), ())
        self.assertEqual(simple.getDomains(), ())

    def test_getRolesInContext(self):
        null = self._makeOne()
        self.assertEqual(null.getRolesInContext(self), ())

    def test_authenticate(self):
        null = self._makeOne()
        self.assertFalse(null.authenticate('password', {}))

    def test_allowed(self):
        null = self._makeOne()
        self.assertFalse(null.allowed(self, ()))

    def test_has_role(self):
        null = self._makeOne()
        self.assertFalse(null.has_role('Authenticated'))

    def test_has_role_w_object(self):
        null = self._makeOne()
        self.assertFalse(null.has_role('Authenticated', self))

    def test_has_permission(self):
        null = self._makeOne()
        self.assertFalse(null.has_permission('View', self))

    def test___repr__(self):
        null = self._makeOne()
        self.assertEqual(repr(null), "<NullUnrestrictedUser (None, None)>")

    def test___str__(self):
        # See https://bugs.launchpad.net/zope2/+bug/142563
        null = self._makeOne()
        self.assertEqual(str(null), "<NullUnrestrictedUser (None, None)>")



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""SecurityInfo objects and access control constants.

   SecurityInfo objects are used in class definitions to allow
   a declarative style of associating access control information
   with class attributes.

   More information on using SecurityInfo and guide to Zope security
   for developers can be found in the dev.zope.org "Declarative Security"
   project:

   http://dev.zope.org/Wikis/DevSite/Projects/DeclarativeSecurity

   While SecurityInfo objects largely remove the need for Python
   programmers to care about the underlying implementation, there
   are several constants defined that should be used by code that
   must set __roles__ attributes directly. (the constants are also
   accessible from the AccessControl namespace). The defined access
   control constants and their meanings are:

   ACCESS_PUBLIC:  accessible from restricted code and possibly
                   through the web (if object has a docstring)

   ACCESS_PRIVATE: accessible only from python code

   ACCESS_NONE:    no access

"""

import sys
from logging import getLogger

from Acquisition import Implicit
from Persistence import Persistent

from AccessControl.class_init import InitializeClass
from AccessControl.ImplPython import _what_not_even_god_should_do


# always patch Persistent before ClassSecurityInfo is used
Persistent.__class_init__ = InitializeClass

LOG = getLogger('SecurityInfo')

# Security constants - these are imported into the AccessControl
# namespace and can be referenced as AccessControl.PUBLIC etc.

ACCESS_NONE = _what_not_even_god_should_do
ACCESS_PRIVATE = ()
ACCESS_PUBLIC = None

_marker = []


class SecurityInfo(Implicit):
    """Encapsulate security information."""

    __security_info__ = 1

    __roles__ = ACCESS_PRIVATE

    def __init__(self):
        self.names = {}
        self.roles = {}
        self._unused_protected_decorators = set()

    def _setaccess(self, names, access):
        for name in names:
            if self.names.get(name, access) != access:
                LOG.warn('Conflicting security declarations for "%s"' % name)
                self._warnings = 1
            self.names[name] = access

    declarePublic__roles__ = ACCESS_PRIVATE

    def declarePublic(self, name, *names):
        """Declare names to be publicly accessible."""
        self._setaccess((name,) + names, ACCESS_PUBLIC)

    declarePrivate__roles__ = ACCESS_PRIVATE

    def declarePrivate(self, name, *names):
        """Declare names to be inaccessible to restricted code."""
        self._setaccess((name,) + names, ACCESS_PRIVATE)

    declareProtected__roles__ = ACCESS_PRIVATE

    def declareProtected(self, permission_name, name, *names):
        """Declare names to be associated with a permission."""
        self._setaccess((name,) + names, permission_name)

    declareObjectPublic__roles__ = ACCESS_PRIVATE

    def declareObjectPublic(self):
        """Declare the object to be publicly accessible."""
        self._setaccess(('',), ACCESS_PUBLIC)

    declareObjectPrivate__roles__ = ACCESS_PRIVATE

    def declareObjectPrivate(self):
        """Declare the object to be inaccessible to restricted code."""
        self._setaccess(('',), ACCESS_PRIVATE)

    declareObjectProtected__roles__ = ACCESS_PRIVATE

    def declareObjectProtected(self, permission_name):
        """Declare the object to be associated with a permission."""
        self._setaccess(('',), permission_name)

    public__roles__ = ACCESS_PRIVATE

    def public(self, func):
        """Decorate a function to be publicly accessible."""
        self.declarePublic(func.__name__)  # NOQA: D001
        return func

    private__roles__ = ACCESS_PRIVATE

    def private(self, func):
        """Decorate a function to be inaccessible to restricted code."""
        self.declarePrivate(func.__name__)  # NOQA: D001
        return func

    protected__roles__ = ACCESS_PRIVATE

    def protected(self, permission_name):
        """Return a decorator to associate a function with a permission."""
        # the decorator returned is remembered in a set and will
        # remove itself upon call. self.apply will check for an empty
        # set and raise an AssertionError otherwise.
        key = "'%s':%s" % (permission_name, id(lambda x: x))

        def decor(func):
            self.declareProtected(permission_name, func.__name__)  # NOQA: D001
            self._unused_protected_decorators.remove(key)
            return func
        # make sure our key algo creates unique-enough keys
        if key in self._unused_protected_decorators:
            raise KeyError("Duplicate key: %s" % (key,))
        self._unused_protected_decorators.add(key)
        return decor

    setPermissionDefault__roles__ = ACCESS_PRIVATE

    def setPermissionDefault(self, permission_name, roles):
        """Declare default roles for a permission"""
        rdict = {}
        for role in roles:
            rdict[role] = 1
        if self.roles.get(permission_name, rdict) != rdict:
            LOG.warn('Conflicting default role'
                     'declarations for permission "%s"' % permission_name)
            self._warnings = 1
        self.roles[permission_name] = rdict

    setDefaultAccess__roles__ = ACCESS_PRIVATE

    def setDefaultAccess(self, access):
        """Declare default attribute access policy.

        This should be a boolean value, a map of attribute names to
        booleans, or a callable (name, value) -> boolean.
        """
        if isinstance(access, str):
            access = access.lower()
            if access == 'allow':
                access = 1
            elif access == 'deny':
                access = 0
            else:
                raise ValueError("'allow' or 'deny' expected")
        self.access = access


class ClassSecurityInfo(SecurityInfo):
    """Encapsulate security information for class objects."""

    __roles__ = ACCESS_PRIVATE

    apply__roles__ = ACCESS_PRIVATE

    def apply(self, classobj):
        """Apply security information to the given class object."""

        # make sure all decorators handed out by security.protected were used
        if self._unused_protected_decorators:
            msg = "Class '%r' has %d non-decorator security.protected calls!"
            raise AssertionError(msg % (classobj,
                                 len(self._unused_protected_decorators)))

        cdict = classobj.__dict__

        # Check the class for an existing __ac_permissions__ and
        # incorporate that if present to support older classes or
        # classes that haven't fully switched to using SecurityInfo.
        if '__ac_permissions__' in cdict:
            for item in cdict['__ac_permissions__']:
                permission_name = item[0]
                self._setaccess(item[1], permission_name)
                if len(item) > 2:
                    self.setPermissionDefault(permission_name, item[2])

        # Set __roles__ for attributes declared public or private.
        # Collect protected attribute names in ac_permissions.
        ac_permissions = {}
        for name, access in self.names.items():
            if access in (ACCESS_PRIVATE, ACCESS_PUBLIC, ACCESS_NONE):
                setattr(classobj, '%s__roles__' % name, access)
            else:
                if access not in ac_permissions:
                    ac_permissions[access] = []
                ac_permissions[access].append(name)

        # Now transform our nested dict structure into the nested tuple
        # structure expected of __ac_permissions__ attributes and set
        # it on the class object.
        getRoles = self.roles.get
        __ac_permissions__ = []
        permissions = sorted(ac_permissions.items())
        for permission_name, names in permissions:
            roles = getRoles(permission_name, ())
            if len(roles):
                entry = (permission_name, tuple(names), tuple(roles.keys()))
            else:
                entry = (permission_name, tuple(names))
            __ac_permissions__.append(entry)
        for permission_name, roles in self.roles.items():
            if permission_name not in ac_permissions:
                entry = (permission_name, (), tuple(roles.keys()))
                __ac_permissions__.append(entry)
        setattr(classobj, '__ac_permissions__', tuple(__ac_permissions__))

        # Take care of default attribute access policy
        access = getattr(self, 'access', _marker)
        if access is not _marker:
            setattr(classobj, '__allow_access_to_unprotected_subobjects__',
                    access)

        if getattr(self, '_warnings', None):
            LOG.warn('Class "%s" had conflicting '
                     'security declarations' % classobj.__name__)


class ClassSecurityInformation(ClassSecurityInfo):
    # Default policy is disallow
    access = 0


_moduleSecurity = {}
_appliedModuleSecurity = {}


def secureModule(mname, *imp):
    modsec = _moduleSecurity.get(mname, None)
    if modsec is None:
        if mname in _appliedModuleSecurity:
            return sys.modules[mname]
        return  # no MSI, no module

    if imp:
        __import__(mname, *imp)
    module = sys.modules[mname]
    modsec.apply(module.__dict__)
    _appliedModuleSecurity[mname] = modsec
    try:
        del _moduleSecurity[mname]
    except KeyError:
        # In case of access from multiple threads, the del might fail, but that
        # is OK.
        pass
    return module


def ModuleSecurityInfo(module_name=None):
    if module_name is not None:
        modsec = _moduleSecurity.get(module_name, None)
        if modsec is not None:
            return modsec
        modsec = _appliedModuleSecurity.get(module_name, None)
        if modsec is not None:
            # Move security info back to to-apply dict (needed for product
            # refresh). Also invoke this check for parent packages already
            # applied
            del _appliedModuleSecurity[module_name]
            _moduleSecurity[module_name] = modsec
            dot = module_name.rfind('.')
            if dot > 0:
                ModuleSecurityInfo(module_name[:dot])
            return modsec
        dot = module_name.rfind('.')
        if dot > 0:
            # If the module is in a package, recursively make sure
            # there are security declarations for the package steps
            # leading to the module
            modname = module_name[dot + 1:]
            pmodsec = ModuleSecurityInfo(module_name[:dot])
            if modname not in pmodsec.names:
                pmodsec.declarePublic(modname)  # NOQA: D001
    return _ModuleSecurityInfo(module_name)


class _ModuleSecurityInfo(SecurityInfo):
    """Encapsulate security information for modules."""

    __roles__ = ACCESS_PRIVATE
    access = False  # deny by default, prevent acquiring 'access' from module

    def __init__(self, module_name=None):
        self.names = {}
        if module_name is not None:
            global _moduleSecurity
            _moduleSecurity[module_name] = self

    __call____roles__ = ACCESS_PRIVATE

    def __call__(self, name, value):
        """Callback for __allow_access_to_unprotected_subobjects__ hook."""
        access = self.names.get(name, _marker)
        if access is not _marker:
            return access == ACCESS_PUBLIC

        return getattr(self, 'access', 0)

    apply__roles__ = ACCESS_PRIVATE

    def apply(self, dict):
        """Apply security information to the given module dict."""

        # Start with default attribute access policy
        access = getattr(self, 'access', _marker)
        if access is not _marker or len(self.names):
            dict['__allow_access_to_unprotected_subobjects__'] = self

        if getattr(self, '_warnings', None):
            LOG.warn('Module "%s" had conflicting '
                     'security declarations' % dict['__name__'])

    declareProtected__roles__ = ACCESS_PRIVATE

    def declareProtected(self, permission_name, *names):
        """Cannot declare module names protected."""
        pass

    declareObjectProtected__roles__ = ACCESS_PRIVATE

    def declareObjectProtected(self, permission_name):
        """Cannot declare module protected."""
        pass

    setDefaultRoles__roles__ = ACCESS_PRIVATE

    def setDefaultRoles(self, permission_name, roles):
        """Cannot set default roles for permissions in a module."""
        pass

# Handy little utility functions


def allow_module(module_name):
    """Allow a module and all its contents to be used from a
    restricted Script. The argument module_name may be a simple
    or dotted module or package name. Note that if a package
    path is given, all modules in the path will be available."""
    ModuleSecurityInfo(module_name).setDefaultAccess(1)
    dot = module_name.find('.')
    while dot > 0:
        ModuleSecurityInfo(module_name[:dot]).setDefaultAccess(1)
        dot = module_name.find('.', dot + 1)


def allow_class(Class):
    """Allow a class and all of its methods to be used from a
    restricted Script.  The argument Class must be a class."""
    Class._security = sec = ClassSecurityInfo()
    sec.declareObjectPublic()
    sec.setDefaultAccess(1)
    sec.apply(Class)
    InitializeClass(Class)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Class initialization.
"""

import logging

from AccessControl.Permission import ApplicationDefaultPermissions  # NOQA


def InitializeClass(self):
    from AccessControl.Permission import registerPermissions
    from AccessControl.PermissionRole import PermissionRole

    dict = self.__dict__
    have = dict.__contains__
    ft = type(InitializeClass)
    dict_items = list(dict.items())

    for name, v in dict_items:
        if getattr(v, '_need__name__', 0):
            d = v.__dict__
            oldname = d.get('__name__', '')
            if d.get('_implicit__name__', 0):
                # Already supplied a name.
                if name != oldname:
                    # Tried to implicitly assign a different name!
                    try:
                        classname = '%s.%s' % (self.__module__, self.__name__)
                    except AttributeError:
                        classname = repr(self)
                    logging.getLogger("Init").warning(
                        'Ambiguous name for method of %s: %r != %r',
                        classname, d['__name__'], name)
            else:
                # Supply a name implicitly so that the method can
                # find the security assertions on its container.
                v._implicit__name__ = 1
                v.__name__ = name
            if name == 'manage' or name[:7] == 'manage_':
                name = name + '__roles__'
                if not have(name):
                    setattr(self, name, ('Manager',))
        elif name == 'manage' or name[:7] == 'manage_' and isinstance(v, ft):
            name = name + '__roles__'
            if not have(name):
                setattr(self, name, ('Manager',))

    # Look for a SecurityInfo object on the class. If found, call its
    # apply() method to generate __ac_permissions__ for the class. We
    # delete the SecurityInfo from the class dict after it has been
    # applied out of paranoia.
    for key, value in dict_items:
        if hasattr(value, '__security_info__'):
            security_info = value
            security_info.apply(self)
            delattr(self, key)
            break

    if '__ac_permissions__' in self.__dict__:
        registerPermissions(self.__ac_permissions__)
        for acp in self.__ac_permissions__:
            pname, mnames = acp[:2]
            if len(acp) > 2:
                roles = acp[2]
                pr = PermissionRole(pname, roles)
            else:
                pr = PermissionRole(pname)
            for mname in mnames:
                setattr(self, mname + '__roles__', pr)
                if mname and \
                   mname not in ('context', 'request') and \
                   not hasattr(self, mname):
                    # don't complain about context or request, as they are
                    # frequently not available as class attributes
                    logging.getLogger("Init").warning(
                        "Class %s.%s has a security declaration for "
                        "nonexistent method %r", self.__module__,
                        self.__name__, mname)


default__class_init__ = InitializeClass  # BBB: old name



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" TaintedString tests.
"""

import unittest

import six


class TestFunctions(unittest.TestCase):

    def test_taint_string(self):
        from AccessControl.tainted import TaintedBytes
        from AccessControl.tainted import TaintedString
        from AccessControl.tainted import taint_string
        self.assertIsInstance(taint_string('string'), TaintedString)
        self.assertIsInstance(taint_string(b'bytes'), TaintedBytes)

    def test_should_be_tainted(self):
        from AccessControl.tainted import should_be_tainted
        self.assertFalse(should_be_tainted('string'))
        self.assertTrue(should_be_tainted('<string'))
        self.assertFalse(should_be_tainted(b'string'))
        self.assertTrue(should_be_tainted(b'<string'))
        self.assertFalse(should_be_tainted(b'string'[0]))
        self.assertTrue(should_be_tainted(b'<string'[0]))


class TestTaintedString(unittest.TestCase):

    def setUp(self):
        self.unquoted = '<test attr="&">'
        self.quoted = '&lt;test attr=&quot;&amp;&quot;&gt;'
        self.tainted = self._getClass()(self.unquoted)

    def _getClass(self):
        from AccessControl.tainted import TaintedString
        return TaintedString

    def testStr(self):
        self.assertEqual(str(self.tainted), self.unquoted)

    def testRepr(self):
        self.assertEqual(repr(self.tainted), repr(self.quoted))

    def testEqual(self):
        self.assertEqual(self.tainted, self.unquoted)

    def testCmp(self):
        self.assertTrue(self.tainted == self.unquoted)
        self.assertTrue(self.tainted < 'a')
        self.assertTrue(self.tainted > '.')

    def testHash(self):
        hash = {}
        hash[self.tainted] = self.quoted
        hash[self.unquoted] = self.unquoted
        self.assertEqual(hash[self.tainted], self.unquoted)

    def testLen(self):
        self.assertEqual(len(self.tainted), len(self.unquoted))

    def testGetItem(self):
        self.assertTrue(isinstance(self.tainted[0], self._getClass()))
        self.assertEqual(self.tainted[0], '<')
        self.assertFalse(isinstance(self.tainted[-1], self._getClass()))
        self.assertEqual(self.tainted[-1], '>')

    def testGetSlice(self):
        self.assertTrue(isinstance(self.tainted[0:1], self._getClass()))
        self.assertEqual(self.tainted[0:1], '<')
        self.assertFalse(isinstance(self.tainted[1:], self._getClass()))
        self.assertEqual(self.tainted[1:], self.unquoted[1:])

    CONCAT = 'test'

    def testConcat(self):
        self.assertTrue(isinstance(self.tainted + self.CONCAT,
                        self._getClass()))
        self.assertEqual(self.tainted + self.CONCAT,
                         self.unquoted + self.CONCAT)
        self.assertTrue(isinstance(self.CONCAT + self.tainted,
                                   self._getClass()))
        self.assertEqual(self.CONCAT + self.tainted,
                         self.CONCAT + self.unquoted)

    def testMultiply(self):
        self.assertTrue(isinstance(2 * self.tainted, self._getClass()))
        self.assertEqual(2 * self.tainted, 2 * self.unquoted)
        self.assertTrue(isinstance(self.tainted * 2, self._getClass()))
        self.assertEqual(self.tainted * 2, self.unquoted * 2)

    def testInterpolate(self):
        tainted = self._getClass()('<%s>')
        self.assertTrue(isinstance(tainted % 'foo', self._getClass()))
        self.assertEqual(tainted % 'foo', '<foo>')
        tainted = self._getClass()('<%s attr="%s">')
        self.assertTrue(isinstance(tainted % ('foo', 'bar'), self._getClass()))
        self.assertEqual(tainted % ('foo', 'bar'), '<foo attr="bar">')

    def testStringMethods(self):
        simple = "capitalize isalpha isdigit islower isspace istitle isupper" \
            " lower lstrip rstrip strip swapcase upper".split()
        returnsTainted = "capitalize lower lstrip rstrip strip swapcase upper"
        returnsTainted = returnsTainted.split()
        unquoted = '\tThis is a test  '
        tainted = self._getClass()(unquoted)
        for f in simple:
            v = getattr(tainted, f)()
            self.assertEqual(v, getattr(unquoted, f)())
            if f in returnsTainted:
                self.assertTrue(isinstance(v, self._getClass()))
            else:
                self.assertFalse(isinstance(v, self._getClass()))

        optArg = "lstrip rstrip strip".split()
        for f in optArg:
            v = getattr(tainted, f)(" ")
            self.assertEqual(v, getattr(unquoted, f)(" "))
            self.assertTrue(isinstance(v, self._getClass()))

        justify = "center ljust rjust".split()
        for f in justify:
            v = getattr(tainted, f)(30)
            self.assertEqual(v, getattr(unquoted, f)(30))
            self.assertTrue(isinstance(v, self._getClass()))

        searches = "find index rfind rindex endswith startswith".split()
        searchraises = "index rindex".split()
        for f in searches:
            v = getattr(tainted, f)('test')
            self.assertEqual(v, getattr(unquoted, f)('test'))
            if f in searchraises:
                self.assertRaises(ValueError, getattr(tainted, f), 'nada')

        self.assertEqual(tainted.count('test', 1, -1),
                         unquoted.count('test', 1, -1))

        self.assertEqual(tainted.encode(), unquoted.encode())
        self.assertTrue(isinstance(tainted.encode(), self._getClass()))

        self.assertEqual(tainted.expandtabs(10), unquoted.expandtabs(10))
        self.assertTrue(isinstance(tainted.expandtabs(), self._getClass()))

        self.assertEqual(tainted.replace('test', 'spam'),
                         unquoted.replace('test', 'spam'))
        self.assertTrue(isinstance(tainted.replace('test', '<'),
                                   self._getClass()))
        self.assertFalse(isinstance(tainted.replace('test', 'spam'),
                                    self._getClass()))

        self.assertEqual(tainted.split(), unquoted.split())
        for part in self._getClass()('< < <').split():
            self.assertTrue(isinstance(part, self._getClass()))
        for part in tainted.split():
            self.assertFalse(isinstance(part, self._getClass()))

        multiline = 'test\n<tainted>'
        lines = self._getClass()(multiline).split()
        self.assertEqual(lines, multiline.split())
        self.assertTrue(isinstance(lines[1], self._getClass()))
        self.assertFalse(isinstance(lines[0], self._getClass()))

        transtable = ''.join(map(chr, range(256)))
        self.assertEqual(tainted.translate(transtable),
                         unquoted.translate(transtable))
        self.assertTrue(isinstance(self._getClass()('<').translate(transtable),
                                   self._getClass()))
        if six.PY2:
            # Translate no longer supports a second argument
            kls = self._getClass()('<')
            self.assertFalse(isinstance(kls.translate(transtable, '<'),
                                        self._getClass()))

    def testQuoted(self):
        self.assertEqual(self.tainted.quoted(), self.quoted)


class TestTaintedBytes(TestTaintedString):

    def setUp(self):
        self.unquoted = b'<test attr="&">'
        self.quoted = b'&lt;test attr=&quot;&amp;&quot;&gt;'
        self.tainted = self._getClass()(self.unquoted)

    def _getClass(self):
        from AccessControl.tainted import TaintedBytes
        return TaintedBytes

    def testCmp(self):
        self.assertTrue(self.tainted == self.unquoted)
        self.assertEqual(self.tainted, self.unquoted)
        self.assertTrue(self.tainted < b'a')
        self.assertTrue(self.tainted > b'.')

    CONCAT = b'test'

    def testGetItem(self):
        self.assertTrue(isinstance(self.tainted[0], self._getClass()))
        self.assertEqual(self.tainted[0], self._getClass()(b'<'))
        self.assertFalse(isinstance(self.tainted[-1], self._getClass()))
        self.assertEqual(self.tainted[-1], 62 if six.PY3 else b'>')

    def testStr(self):
        self.assertEqual(str(self.tainted), self.unquoted.decode('utf8'))

    def testGetSlice(self):
        self.assertTrue(isinstance(self.tainted[0:1], self._getClass()))
        self.assertEqual(self.tainted[0:1], b'<')
        self.assertFalse(isinstance(self.tainted[1:], self._getClass()))
        self.assertEqual(self.tainted[1:], self.unquoted[1:])

    def testInterpolate(self):
        tainted = self._getClass()(b'<%s>')
        self.assertTrue(isinstance(tainted % b'foo', self._getClass()))
        self.assertEqual(tainted % b'foo', b'<foo>')
        tainted = self._getClass()(b'<%s attr="%s">')
        self.assertTrue(isinstance(tainted % (b'foo', b'bar'),
                                   self._getClass()))
        self.assertEqual(tainted % (b'foo', b'bar'), b'<foo attr="bar">')

    def testStringMethods(self):
        simple = "capitalize isalpha isdigit islower isspace istitle isupper" \
            " lower lstrip rstrip strip swapcase upper".split()
        returnsTainted = "capitalize lower lstrip rstrip strip swapcase upper"
        returnsTainted = returnsTainted.split()
        unquoted = b'\tThis is a test  '
        tainted = self._getClass()(unquoted)
        for f in simple:
            v = getattr(tainted, f)()
            self.assertEqual(v, getattr(unquoted, f)())
            if f in returnsTainted:
                self.assertTrue(isinstance(v, self._getClass()))
            else:
                self.assertFalse(isinstance(v, self._getClass()))

        optArg = "lstrip rstrip strip".split()
        for f in optArg:
            v = getattr(tainted, f)(b" ")
            self.assertEqual(v, getattr(unquoted, f)(b" "))
            self.assertTrue(isinstance(v, self._getClass()))

        justify = "center ljust rjust".split()
        for f in justify:
            v = getattr(tainted, f)(30)
            self.assertEqual(v, getattr(unquoted, f)(30))
            self.assertTrue(isinstance(v, self._getClass()))

        searches = "find index rfind rindex endswith startswith".split()
        searchraises = "index rindex".split()
        for f in searches:
            v = getattr(tainted, f)(b'test')
            self.assertEqual(v, getattr(unquoted, f)(b'test'))
            if f in searchraises:
                self.assertRaises(ValueError, getattr(tainted, f), b'nada')

        self.assertEqual(tainted.count(b'test', 1, -1),
                         unquoted.count(b'test', 1, -1))

        self.assertEqual(tainted.decode(), unquoted.decode())
        from AccessControl.tainted import TaintedString
        self.assertTrue(isinstance(tainted.decode(), TaintedString))

        self.assertEqual(tainted.expandtabs(10), unquoted.expandtabs(10))
        self.assertTrue(isinstance(tainted.expandtabs(), self._getClass()))

        self.assertEqual(tainted.replace(b'test', b'spam'),
                         unquoted.replace(b'test', b'spam'))
        self.assertTrue(isinstance(tainted.replace(b'test', b'<'),
                                   self._getClass()))
        self.assertFalse(isinstance(tainted.replace(b'test', b'spam'),
                                    self._getClass()))

        self.assertEqual(tainted.split(), unquoted.split())
        for part in self._getClass()(b'< < <').split():
            self.assertTrue(isinstance(part, self._getClass()))
        for part in tainted.split():
            self.assertFalse(isinstance(part, self._getClass()))

        multiline = b'test\n<tainted>'
        lines = self._getClass()(multiline).split()
        self.assertEqual(lines, multiline.split())
        self.assertTrue(isinstance(lines[1], self._getClass()))
        self.assertFalse(isinstance(lines[0], self._getClass()))

        if six.PY3:
            transtable = bytes(range(256))
        else:
            transtable = ''.join(map(chr, range(256)))
        self.assertEqual(tainted.translate(transtable),
                         unquoted.translate(transtable))
        kls = self._getClass()(b'<')
        self.assertTrue(isinstance(kls.translate(transtable),
                                   self._getClass()))
        if six.PY2:
            # Translate no longer supports a second argument
            kls = self._getClass()(b'<')
            self.assertFalse(isinstance(kls.translate(transtable, b'<'),
                                        self._getClass()))

    def testConstructor(self):
        from AccessControl.tainted import TaintedBytes
        if six.PY2:
            self.assertRaises(ValueError, TaintedBytes, [60])
        if six.PY3:
            self.assertEqual(TaintedBytes(60), b'<')
            self.assertEqual(TaintedBytes(32), b' ')
        self.assertEqual(TaintedBytes(b'abc'), b'abc')
        self.assertRaises(ValueError, TaintedBytes, u"abc")



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
'''Collect rules for access to objects that don\'t have roles.

The rules are expressed as a mapping from type -> assertion

An assertion can be:

  - A dict

  - A callable

  - Something with a truth value

If the assertion is a callable, then it will be called with
a name being accessed and the name used.  Its return value is ignored,
but in may veto an access by raising an exception.

If the assertion is a dictionary, then the keys are attribute names.
The values may be callables or objects with boolean values. If a value
is callable, it will be called with the object we are accessing an
attribute of and the attribute name. It should return an attribute
value. Callables are often used to returned guarded versions of
methods.  Otherwise, accesses are allowed if values in this dictionary
are true and disallowed if the values are false or if an item for an
attribute name is not present.

If the assertion is not a dict and is not callable, then access to
unprotected attributes is allowed if the assertion is true, and
disallowed otherwise.

XXX This descrition doesn't actually match what's done in ZopeGuards
or in ZopeSecurityPolicy. :(
'''

from contextlib import contextmanager

from BTrees.IIBTree import IIBTree
from BTrees.IIBTree import IIBucket
from BTrees.IIBTree import IISet
from BTrees.IOBTree import IOBTree
from BTrees.IOBTree import IOBucket
from BTrees.IOBTree import IOSet
from BTrees.OIBTree import OIBTree
from BTrees.OIBTree import OIBucket
from BTrees.OIBTree import OISet
#
#   WAAAA!
#
from BTrees.OOBTree import OOBTree
from BTrees.OOBTree import OOBucket
from BTrees.OOBTree import OOSet


_noroles = []  # this is imported in various places
_marker = object()

# ContainerAssertions are used by cAccessControl to check access to
# attributes of container types, like dict, list, or string.
# ContainerAssertions maps types to a either a dict, a function, or a
# simple boolean value.  When guarded_getattr checks the type of its
# first argument against ContainerAssertions, and invokes checking
# logic depending on what value it finds.

# If the value for a type is:
#   - a boolean value:
#      - the value determines whether access is allowed
#   - a function (or callable):
#      - The function is called with the name of the attribute and
#        the actual attribute value, then the value is returned.
#        The function can raise an exception.
#   - a dict:
#      - The dict maps attribute names to boolean values or functions.
#        The boolean values behave as above, but the functions do not.
#        The value returned for attribute access is the result of
#        calling the function with the object and the attribute name.
ContainerAssertions = {
    type(()): 1,
    type(b''): 1,
    type(u''): 1,
    range: 1,
}

Containers = ContainerAssertions.get


def allow_type(Type, allowed=1):
    """Allow a type and all of its methods and attributes to be used from
    restricted code.  The argument Type must be a type."""
    if not isinstance(Type, type):
        raise ValueError("%r is not a type" % Type)
    if hasattr(Type, '__roles__'):
        raise ValueError("%r handles its own security" % Type)
    if not (isinstance(allowed, int) or isinstance(allowed, dict)):
        raise ValueError("The 'allowed' argument must be an int or dict.")
    ContainerAssertions[Type] = allowed


for tree_type, has_values in [
    (OOBTree, 1),
    (OOBucket, 1),
    (OOSet, 0),
    (OIBTree, 1),
    (OIBucket, 1),
    (OISet, 0),
    (IOBTree, 1),
    (IOBucket, 1),
    (IOSet, 0),
    (IIBTree, 1),
    (IIBucket, 1),
    (IISet, 0),
]:
    tree = tree_type()
    keys = tree.keys()

    if not isinstance(keys, list):  # lists have their own declarations
        allow_type(type(keys))

    if has_values:
        assert isinstance(keys, type(tree.values()))
        assert isinstance(keys, type(tree.items()))


@contextmanager
def override_containers(type_, assertions):
    """Temporarily override the container assertions."""
    orig_container = Containers(type_, _marker)
    ContainerAssertions[type_] = assertions
    try:
        yield
    finally:
        if orig_container is _marker:
            del ContainerAssertions[type_]
        else:
            ContainerAssertions[type_] = orig_container



#############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import six

from zExceptions import Forbidden
from zope.publisher.interfaces.browser import IBrowserRequest


if six.PY3:
    from inspect import getfullargspec
    from inspect import signature
else:  # Python 2
    from inspect import getargspec as getfullargspec

    from funcsigs import signature

_default = []


def buildfacade(name, method, docstring):
    """Build a facade function, matching the decorated method in signature.

    Note that defaults are replaced by _default, and _curried will reconstruct
    these to preserve mutable defaults.

    """
    sig = signature(method)
    args = []
    callargs = []
    for v in sig.parameters.values():
        parts = str(v).split('=')
        args.append(
            parts[0] if len(parts) == 1 else
            '{}=_default'.format(parts[0]))  # NOQA: 43
        callargs.append(parts[0])
    return 'def %s(%s):\n    """%s"""\n    return _curried(%s)' % (
        name, ', '.join(args), docstring, ', '.join(callargs))


def requestmethod(*methods):
    """Create a request method specific decorator"""
    methods = list(map(lambda m: m.upper(), methods))
    if len(methods) > 1:
        methodsstr = ', '.join(methods[:-1])
        methodsstr += ' or ' + methods[-1]
    else:
        methodsstr = methods[0]

    def _methodtest(callable):
        """Only allow callable when request method is %s.""" % methodsstr
        spec = getfullargspec(callable)
        args, defaults = spec[0], spec[3]
        try:
            r_index = args.index('REQUEST')
        except ValueError:
            raise ValueError('No REQUEST parameter in callable signature')

        arglen = len(args)
        if defaults is not None:
            defaults = list(zip(args[arglen - len(defaults):], defaults))
            arglen -= len(defaults)

        def _curried(*args, **kw):
            request = args[r_index]
            if IBrowserRequest.providedBy(request):
                if request.method not in methods:
                    raise Forbidden('Request must be %s' % methodsstr)

            # Reconstruct keyword arguments
            if defaults is not None:
                args, kwparams = args[:arglen], args[arglen:]
                for positional, (key, default) in zip(kwparams, defaults):
                    if positional is _default:
                        kw[key] = default
                    else:
                        kw[key] = positional

            return callable(*args, **kw)

        # Build a facade, with a reference to our locally-scoped _curried
        name = callable.__name__
        facade_globs = dict(_curried=_curried, _default=_default)
        exec(buildfacade(name, callable, callable.__doc__), facade_globs)
        return facade_globs[name]

    return _methodtest


# For Zope versions 2.8 - 2.10
postonly = requestmethod('POST')

__all__ = ('requestmethod', 'postonly')



import string

import six

from AccessControl.ZopeGuards import guarded_getattr
from AccessControl.ZopeGuards import guarded_getitem


try:
    # Python 3
    import _string
except ImportError:
    pass

if six.PY3:
    from collections.abc import Mapping
else:
    from collections import Mapping


def formatter_field_name_split(field_name):
    if six.PY3:
        return _string.formatter_field_name_split(field_name)
    else:
        return field_name._formatter_field_name_split()


class _MagicFormatMapping(Mapping):
    """Pulled from Jinja2.

    This class implements a dummy wrapper to fix a bug in the Python
    standard library for string formatting.

    See http://bugs.python.org/issue13598 for information about why
    this is necessary.
    """

    def __init__(self, args, kwargs):
        self._args = args
        self._kwargs = kwargs
        self._last_index = 0

    def __getitem__(self, key):
        if key == '':
            idx = self._last_index
            self._last_index += 1
            try:
                return self._args[idx]
            except LookupError:
                pass
            key = str(idx)
        return self._kwargs[key]

    def __iter__(self):
        return iter(self._kwargs)

    def __len__(self):
        return len(self._kwargs)


class SafeFormatter(string.Formatter):
    """Formatter using guarded access."""

    def __init__(self, value):
        self.value = value
        super(SafeFormatter, self).__init__()

    def get_field(self, field_name, args, kwargs):
        """Get the field value using guarded methods."""
        first, rest = formatter_field_name_split(field_name)

        obj = self.get_value(first, args, kwargs)

        # loop through the rest of the field_name, doing
        #  getattr or getitem as needed
        for is_attr, i in rest:
            if is_attr:
                obj = guarded_getattr(obj, i)
            else:
                obj = guarded_getitem(obj, i)

        return obj, first

    def safe_format(self, *args, **kwargs):
        """Safe variant of `format` method."""
        kwargs = _MagicFormatMapping(args, kwargs)
        return self.vformat(self.value, args, kwargs)


def safe_format(inst, method):
    """Use our SafeFormatter that uses guarded_getattr for attribute access."""
    return SafeFormatter(inst).safe_format



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""C implementation of the access control machinery."""

from AccessControl.cAccessControl import PermissionRole
from AccessControl.cAccessControl import SecurityManager as cSecurityManager
from AccessControl.cAccessControl import \
    ZopeSecurityPolicy as cZopeSecurityPolicy
from AccessControl.cAccessControl import _what_not_even_god_should_do
from AccessControl.cAccessControl import aq_validate
from AccessControl.cAccessControl import guarded_getattr
from AccessControl.cAccessControl import imPermissionRole
from AccessControl.cAccessControl import rolesForPermissionOn
from AccessControl.cAccessControl import setDefaultBehaviors
from AccessControl.ImplPython import SecurityManager
from AccessControl.ImplPython import ZopeSecurityPolicy


class ZopeSecurityPolicy(cZopeSecurityPolicy, ZopeSecurityPolicy):
    """A security manager provides methods for checking access and managing
    executable context and policies
    """


class SecurityManager(cSecurityManager, SecurityManager):
    """A security manager provides methods for checking access and managing
    executable context and policies
    """



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import unittest


class ClassSecurityInfoTests(unittest.TestCase):

    def _getTargetClass(self):
        from AccessControl.SecurityInfo import ClassSecurityInfo
        return ClassSecurityInfo

    def assertRaises(self, excClass, callableObj, *args, **kwargs):
        """Fail unless an exception of class excClass is thrown
           by callableObj when invoked with arguments args and keyword
           arguments kwargs. If a different type of exception is
           thrown, it will not be caught, and the test case will be
           deemed to have suffered an error, exactly as for an
           unexpected exception.

           Return the raised exception object, if it matches the expected type.
        """
        try:
            callableObj(*args, **kwargs)
        except excClass as e:
            return e
        else:
            if getattr(excClass, '__name__', None) is not None:
                excName = excClass.__name__
            else:
                excName = str(excClass)
            raise self.failureException("%s not raised" % excName)

    def test_SetPermissionDefault(self):
        # Test setting default roles for permissions.
        from ExtensionClass import Base

        from AccessControl.class_init import InitializeClass

        ClassSecurityInfo = self._getTargetClass()

        # Setup a test class with default role -> permission decls.
        class Test(Base):
            """Test class
            """
            __ac_roles__ = ('Role A', 'Role B', 'Role C')
            meta_type = "Test"
            security = ClassSecurityInfo()
            security.setPermissionDefault('Make food', ('Chef', ))
            security.setPermissionDefault(
                'Test permission',
                ('Manager', 'Role A', 'Role B', 'Role C'),
            )

            security.declarePublic('public')  # NOQA: D001

            def public(self, REQUEST=None):
                """ """

            security.declarePrivate('private')  # NOQA: D001

            def private(self, REQUEST=None):
                """ """

            security.declareProtected('Test permission', 'protected')   # NOQA

            def protected(self, REQUEST=None):
                """ """

            # same with decorators

            @security.public
            def public_new(self, REQUEST=None):
                """ """

            @security.private
            def private_new(self, REQUEST=None):
                """ """

            @security.protected('Test permission')
            def protected_new(self, REQUEST=None):
                """ """

        # Do class initialization.
        InitializeClass(Test)

        # Now check the resulting class to see if the mapping was made
        # correctly. Note that this uses carnal knowledge of the internal
        # structures used to store this information!
        object = Test()
        self.assertEqual(object.public__roles__, None)
        self.assertEqual(object.private__roles__, ())
        imPermissionRole = [r for r in object.protected__roles__
                            if not r.endswith('_Permission')]
        self.assertTrue(len(imPermissionRole) == 4)

        for item in ('Manager', 'Role A', 'Role B', 'Role C'):
            self.assertTrue(item in imPermissionRole)

        # functions exist, i.e. decorators returned them
        self.assertEqual(object.public_new.__name__, 'public_new')
        self.assertEqual(object.private_new.__name__, 'private_new')
        self.assertEqual(object.protected_new.__name__, 'protected_new')

        # roles for functions have been set via decorators
        self.assertEqual(object.public_new__roles__, None)
        self.assertEqual(object.private_new__roles__, ())
        imPermissionRole = [r for r in object.protected_new__roles__
                            if not r.endswith('_Permission')]
        self.assertTrue(len(imPermissionRole) == 4)

        for item in ('Manager', 'Role A', 'Role B', 'Role C'):
            self.assertTrue(item in imPermissionRole)

        # Make sure that a permission defined without accompanying method
        # is still reflected in __ac_permissions__
        self.assertEqual([t for t in Test.__ac_permissions__ if not t[1]],
                         [('Make food', (), ('Chef',))])

    def test_EnsureProtectedDecoCall(self):
        from ExtensionClass import Base

        from AccessControl.class_init import InitializeClass

        ClassSecurityInfo = self._getTargetClass()

        class Test(Base):
            """Test class
            """
            meta_type = "Test"

            security = ClassSecurityInfo()

            # security not used as a decorator, so does not protect
            security.protected('Test permission 1')

            def unprotected1(self, REQUEST=None):
                """ """

            # see above
            security.protected('Test permission 2')

            def unprotected2(self, REQUEST=None):
                """ """

            @security.protected('Test permission 3')
            def protected(self, REQUEST=None):
                """ """

        # Do class initialization.
        exc = self.assertRaises(AssertionError,
                                InitializeClass, Test)
        self.assertTrue('has 2 non-decorator' in str(exc))

    def test_aq_context_in_decorators(self):
        from Acquisition import Implicit
        info = self._getTargetClass()

        class A(Implicit):
            security = info()
            a = 1

            @security.public
            def public(self):
                return self.a

            @security.private
            def private(self):
                # make sure the acquisition context is still intact
                return self.b

        class B(Implicit):
            security = info()
            b = 2

        a = A()
        b = B()
        a = a.__of__(b)

        self.assertEqual(a.b, 2)
        self.assertEqual(a.public(), 1)
        self.assertEqual(a.private(), 2)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Permissions
"""

import re

from Acquisition import aq_base


_NOT_ALLOWED_CHARS = re.compile(r'[^a-zA-Z0-9_]')


def getPermissionIdentifier(name):
    return '_' + re.sub(_NOT_ALLOWED_CHARS, '_', name) + "_Permission"


# BBB
pname = getPermissionIdentifier

_marker = []


class Permission:
    # A Permission maps a named logical permission to a set
    # of attribute names. Attribute names which appear in a
    # permission may not appear in any other permission defined
    # by the object.

    def __init__(self, name, data, obj, default=None):
        self.name = name
        self._p = getPermissionIdentifier(name)
        self.data = data
        self.obj = aq_base(obj)
        self.default = default

    def getRoles(self, default=_marker):
        # Return the list of role names which have been given
        # this permission for the object in question. To do
        # this, we try to get __roles__ from all of the object
        # attributes that this permission represents.
        obj = self.obj
        name = self._p
        if hasattr(obj, name):
            return getattr(obj, name)
        roles = default
        for name in self.data:
            if name:
                if hasattr(obj, name):
                    attr = getattr(obj, name)
                    if getattr(attr, '__self__', None) is not None:
                        attr = attr.__self__
                        if hasattr(attr, '__dict__'):
                            attr = attr.__dict__
                            name = name + '__roles__'
                            if name in attr:
                                roles = attr[name]
                                break
            elif hasattr(obj, '__dict__'):
                attr = obj.__dict__
                if '__roles__' in attr:
                    roles = attr['__roles__']
                    break

        if roles:
            try:
                if 'Shared' not in roles:
                    return tuple(roles)
                roles = list(roles)
                roles.remove('Shared')
                return roles
            except BaseException:
                return []

        if roles is None:
            return ['Manager', 'Anonymous']
        if roles is _marker:
            return ['Manager']

        return roles

    def setRoles(self, roles):
        obj = self.obj

        if isinstance(roles, list) and not roles:
            if hasattr(obj, self._p):
                delattr(obj, self._p)
        else:
            setattr(obj, self._p, roles)

        for name in self.data:
            if name == '':
                attr = obj
            else:
                attr = getattr(obj, name)
            try:
                del attr.__roles__
            except BaseException:
                pass
            try:
                delattr(obj, name + '__roles__')
            except BaseException:
                pass

    def setRole(self, role, present):
        roles = self.getRoles()
        if role in roles:
            if present:
                return
            if isinstance(roles, list):
                roles.remove(role)
            else:
                roles = list(roles)
                roles.remove(role)
                roles = tuple(roles)
        elif not present:
            return
        else:
            if isinstance(roles, list):
                roles.append(role)
            else:
                roles = roles + (role, )
        self.setRoles(roles)

    def __len__(self):
        return 1

    def __str__(self):
        return self.name


_registeredPermissions = {}
_ac_permissions = ()


def getPermissions():
    return _ac_permissions


def addPermission(perm, default_roles=('Manager', )):
    if perm in _registeredPermissions:
        return

    entry = ((perm, (), default_roles), )
    global _ac_permissions
    _ac_permissions = _ac_permissions + entry
    _registeredPermissions[perm] = 1
    mangled = getPermissionIdentifier(perm)
    if not hasattr(ApplicationDefaultPermissions, mangled):
        setattr(ApplicationDefaultPermissions, mangled, default_roles)


def registerPermissions(permissions, defaultDefault=('Manager', )):
    """Register an __ac_permissions__ sequence.
    """
    for setting in permissions:
        if setting[0] in _registeredPermissions:
            continue
        if len(setting) == 2:
            perm, methods = setting
            default = defaultDefault
        else:
            perm, methods, default = setting
        addPermission(perm, default)


class ApplicationDefaultPermissions:
    _View_Permission = ('Manager', 'Anonymous')
    _Access_contents_information_Permission = ('Manager', 'Anonymous')



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Constant definitions for built-in Zope permissions
"""

access_contents_information = 'Access contents information'
add_database_methods = 'Add Database Methods'
add_documents_images_and_files = 'Add Documents, Images, and Files'
add_external_methods = 'Add External Methods'
add_folders = 'Add Folders'
add_mailhost_objects = 'Add MailHost objects'
add_page_templates = 'Add Page Templates'
add_python_scripts = 'Add Python Scripts'
add_user_folders = 'Add User Folders'
add_vocabularies = 'Add Vocabularies'
add_z_gadfly_database_connections = 'Add Z Gadfly Database Connections'
add_zcatalogs = 'Add ZCatalogs'
add_zope_tutorials = 'Add Zope Tutorials'
change_database_connections = 'Change Database Connections'
change_database_methods = 'Change Database Methods'
change_external_methods = 'Change External Methods'
change_images_and_files = 'Change Images and Files'
change_python_scripts = 'Change Python Scripts'
change_configuration = 'Change configuration'
change_page_templates = 'Change Page Templates'
change_permissions = 'Change permissions'
change_proxy_roles = 'Change proxy roles'
copy_or_move = 'Copy or Move'
create_class_instances = 'Create class instances'
define_permissions = 'Define permissions'
delete_objects = 'Delete objects'
edit_factories = 'Edit Factories'
ftp_access = 'FTP access'
import_export_objects = 'Import/Export objects'
manage_vocabulary = 'Manage Vocabulary'
manage_zcatalog_entries = 'Manage ZCatalog Entries'
manage_zcatalog_indexes = 'Manage ZCatalogIndex Entries'
manage_properties = 'Manage properties'
manage_users = 'Manage users'
open_close_database_connection = 'Open/Close Database Connection'
open_close_database_connections = 'Open/Close Database Connections'
query_vocabulary = 'Query Vocabulary'
search_zcatalog = 'Search ZCatalog'
take_ownership = 'Take ownership'
test_database_connections = 'Test Database Connections'
undo_changes = 'Undo changes'
use_database_methods = 'Use Database Methods'
use_factories = 'Use Factories'
use_mailhost_services = 'Use mailhost services'
view = 'View'
view_management_screens = 'View management screens'
webdav_access = 'WebDAV access'
webdav_lock_items = 'WebDAV Lock items'
webdav_manage_locks = 'Manage WebDAV Locks'
webdav_unlock_items = 'WebDAV Unlock items'



#############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import unittest

import zExceptions
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserRequest

from AccessControl.requestmethod import buildfacade
from AccessControl.requestmethod import getfullargspec
from AccessControl.requestmethod import requestmethod


@implementer(IBrowserRequest)
class DummyRequest:

    def __init__(self, method):
        self.method = method


GET = DummyRequest('GET')
POST = DummyRequest('POST')


class RequestMethodDecoratorsTests(unittest.TestCase):
    """Using request method decorators, you can limit functions or methods to
    only be callable when the HTTP request was made using a particular method.
    """

    def test_requires_the_defined_HTTP_mehtod(self):
        # To limit access to a function or method to POST requests, use the
        # requestmethod decorator factory:
        @requestmethod('POST')
        def foo(bar, REQUEST):
            return bar

        # When this method is accessed through a request that does not use
        # POST, the Forbidden exception will be raised:
        with self.assertRaises(zExceptions.Forbidden) as err:
            foo('spam', GET)
        self.assertEqual('Request must be POST', str(err.exception))

        # Only when the request was made using POST, will the call succeed:
        self.assertEqual('spam', foo('spam', POST))

    def test_it_does_not_matter_if_REQUEST_is_positional_or_keyword_arg(self):
        # It doesn't matter if REQUEST is a positional or a keyword parameter:

        @requestmethod('POST')
        def foo(bar, REQUEST=None):
            return bar

        with self.assertRaises(zExceptions.Forbidden) as err:
            foo('spam', GET)
        self.assertEqual('Request must be POST', str(err.exception))

        # *Not* passing an optional REQUEST always succeeds::
        self.assertEqual('spam', foo('spam', POST))

    def test_REQUEST_parameter_is_a_requirement(self):
        # Note that the REQUEST parameter is a requirement for the decorator to
        # operate, not including it in the callable signature results in an
        # error:
        try:
            @requestmethod('POST')
            def foo(bar):
                return bar
        except ValueError as e:
            self.assertEqual(
                'No REQUEST parameter in callable signature', str(e))
        else:
            self.fail('Did not raise')

    def test_preserves_keyword_parameter_defaults(self):
        # Because the Zope Publisher uses introspection to match REQUEST
        # variables against callable signatures, the result of the decorator
        # must match the original closely, and keyword parameter defaults must
        # be preserved:
        mutabledefault = dict()

        @requestmethod('POST')
        def foo(bar, baz=mutabledefault, egg=mutabledefault, REQUEST=None,
                *args):
            return bar, baz is mutabledefault, egg is None, REQUEST
        self.assertEqual((['bar', 'baz', 'egg', 'REQUEST'], 'args', None),
                         getfullargspec(foo)[:3])
        self.assertEqual(('spam', True, True, None), foo('spam', egg=None))
        with self.assertRaises(TypeError) as err:
            foo(monty='python')
        self.assertEqual("foo() got an unexpected keyword argument 'monty'",
                         str(err.exception))

    def test_can_be_used_for_any_request_method(self):
        # The `requestmethod` decorator factory can be used for any request
        # method, simply pass in the desired request method:
        @requestmethod('PUT')
        def foo(bar, REQUEST=None):
            return bar

        with self.assertRaises(zExceptions.Forbidden) as err:
            foo('spam', GET)
        self.assertEqual('Request must be PUT', str(err.exception))

    def test_allows_multiple_request_methods(self):
        # You can pass in multiple request methods allow access by any of them:
        @requestmethod('GET', 'HEAD', 'PROPFIND')
        def foo(bar, REQUEST=None):
            return bar

        self.assertEqual('spam', foo('spam', GET))
        with self.assertRaises(zExceptions.Forbidden) as err:
            foo('spam', POST)
        self.assertEqual('Request must be GET, HEAD or PROPFIND',
                         str(err.exception))

    def test_facade_render_correct_args_kwargs(self):
        """ s. https://github.com/zopefoundation/AccessControl/issues/69
        """
        def foo(bar, baz, *args, **kwargs):
            """foo doc"""
            return baz
        got = buildfacade('foo', foo, foo.__doc__)
        expected = '\n'.join([
            'def foo(bar, baz, *args, **kwargs):',
            '    """foo doc"""',
            '    return _curried(bar, baz, *args, **kwargs)'])
        self.assertEqual(expected, got)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Tests of PermissionRole
"""

import unittest

from Acquisition import Explicit
from Acquisition import Implicit
from Acquisition import aq_base

from AccessControl.PermissionRole import PermissionRole


ViewPermission = 'View'
EditThingsPermission = 'Edit Things!'
DeletePermission = 'Delete'


class AppRoot(Explicit):
    _View_Permission = None
    _Edit_Things__Permission = ('Manager', 'Owner')
    # No default for delete permission.


class ImplicitContainer(Implicit):
    pass


class ExplicitContainer(Explicit):
    pass


class RestrictiveObject(Implicit):
    _View_Permission = ('Manager',)
    _Delete_Permission = ()  # Nobody


class PermissiveObject(Explicit):
    _Edit_Things__Permission = ['Anonymous']


def assertPRoles(ob, permission, expect):
    """
    Asserts that in the context of ob, the given permission maps to
    the given roles.
    """
    pr = PermissionRole(permission)
    roles = pr.__of__(ob)
    roles2 = aq_base(pr).__of__(ob)
    assert roles == roles2 or tuple(roles) == tuple(roles2), (
        'Different methods of checking roles computed unequal results')
    same = 0
    if roles:
        # When verbose security is enabled, permission names are
        # embedded in the computed roles.  Remove the permission
        # names.
        roles = [r for r in roles if not r.endswith('_Permission')]

    if roles is None or expect is None:
        if (roles is None or tuple(roles) == ('Anonymous', )) and \
                (expect is None or tuple(expect) == ('Anonymous', )):
            same = 1
    else:
        got = {}
        for r in roles:
            got[r] = 1
        expected = {}
        for r in expect:
            expected[r] = 1
        if got == expected:  # Dict compare does the Right Thing.
            same = 1
    assert same, 'Expected roles: %r, got: %r' % (expect, roles)


class PermissionRoleTests (unittest.TestCase):

    def testRestrictive(self, explicit=0):
        app = AppRoot()
        if explicit:
            app.c = ExplicitContainer()
        else:
            app.c = ImplicitContainer()
        app.c.o = RestrictiveObject()
        o = app.c.o
        assertPRoles(o, ViewPermission, ('Manager', ))
        assertPRoles(o, EditThingsPermission, ('Manager', 'Owner'))
        assertPRoles(o, DeletePermission, ())

    def testPermissive(self, explicit=0):
        app = AppRoot()
        if explicit:
            app.c = ExplicitContainer()
        else:
            app.c = ImplicitContainer()
        app.c.o = PermissiveObject()
        o = app.c.o
        assertPRoles(o, ViewPermission, ('Anonymous', ))
        assertPRoles(o, EditThingsPermission, ('Anonymous',
                                               'Manager',
                                               'Owner'))
        assertPRoles(o, DeletePermission, ('Manager', ))

    def testExplicit(self):
        self.testRestrictive(1)
        self.testPermissive(1)

    def testAppDefaults(self):
        o = AppRoot()
        assertPRoles(o, ViewPermission, ('Anonymous', ))
        assertPRoles(o, EditThingsPermission, ('Manager', 'Owner'))
        assertPRoles(o, DeletePermission, ('Manager', ))

    def testPermissionRoleSupportsGetattr(self):
        a = PermissionRole('a')
        self.assertTrue(getattr(a, '__roles__') == ('Manager', ))
        self.assertTrue(getattr(a, '_d') == ('Manager', ))
        self.assertTrue(getattr(a, '__name__') == 'a')
        self.assertTrue(getattr(a, '_p') == '_a_Permission')



##############################################################################
#
# Copyright (c) 2004-2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Security handling
"""

import six

from zope.component import getUtility
from zope.component import queryUtility
from zope.component.zcml import utility
from zope.configuration.config import GroupingContextDecorator
from zope.configuration.interfaces import IConfigurationContext
from zope.interface import Interface
from zope.interface import implementer
from zope.interface import provider
from zope.schema import ASCIILine
from zope.security.checker import CheckerPublic
from zope.security.interfaces import IInteraction
from zope.security.interfaces import IPermission
from zope.security.interfaces import ISecurityPolicy
from zope.security.management import thread_local
from zope.security.permission import Permission
from zope.security.simplepolicies import ParanoidSecurityPolicy
from zope.security.zcml import IPermissionDirective

from AccessControl.Permission import addPermission
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager


CheckerPublicId = 'zope.Public'
CheckerPrivateId = 'zope2.Private'


def getSecurityInfo(klass):
    sec = {}
    info = vars(klass)
    if '__ac_permissions__' in info:
        sec['__ac_permissions__'] = info['__ac_permissions__']
    for k, v in info.items():
        if k.endswith('__roles__'):
            sec[k] = v
    return sec


def clearSecurityInfo(klass):
    info = vars(klass)
    if '__ac_permissions__' in info:
        delattr(klass, '__ac_permissions__')
    for k, v in list(info.items()):
        if k.endswith('__roles__'):
            delattr(klass, k)


def checkPermission(permission, object, interaction=None):
    """Return whether security policy allows permission on object.

    Arguments:
    permission -- A permission name
    object -- The object being accessed according to the permission
    interaction -- This zope.security concept has no equivalent in Zope 2,
        and is ignored.

    checkPermission is guaranteed to return True if permission is
    CheckerPublic or None.
    """
    if permission in ('zope.Public', 'zope2.Public') or \
       permission is None or \
       permission is CheckerPublic:
        return True

    if isinstance(permission, six.string_types):
        permission = queryUtility(IPermission, six.u(permission))
        if permission is None:
            return False

    if getSecurityManager().checkPermission(permission.title, object):
        return True

    return False


@implementer(IInteraction)
@provider(ISecurityPolicy)
class SecurityPolicy(ParanoidSecurityPolicy):
    """Security policy that bridges between zope.security security mechanisms
    and Zope 2's security policy.

    Don't let the name of the base class fool you... This really just
    delegates to Zope 2's security manager."""

    def checkPermission(self, permission, object):
        return checkPermission(permission, object)


def newInteraction():
    """Con zope.security to use Zope 2's checkPermission.

    zope.security when it does a checkPermission will turn around and
    ask the thread local interaction for the checkPermission method.
    By making the interaction *be* Zope 2's security manager, we can
    con zope.security into using Zope 2's checker...
    """
    if getattr(thread_local, 'interaction', None) is None:
        thread_local.interaction = SecurityPolicy()


def _getSecurity(klass):
    # a Zope 2 class can contain some attribute that is an instance
    # of ClassSecurityInfo. Zope 2 scans through things looking for
    # an attribute that has the name __security_info__ first
    info = vars(klass)
    for k, v in info.items():
        if hasattr(v, '__security_info__'):
            return v
    # we stuff the name ourselves as __security__, not security, as this
    # could theoretically lead to name clashes, and doesn't matter for
    # zope 2 anyway.
    security = ClassSecurityInfo()
    setattr(klass, '__security__', security)
    return security


def protectName(klass, name, permission_id, override_existing_protection=True):
    """Protect the attribute 'name' on 'klass' using the given
       permission"""
    security = _getSecurity(klass)
    # Zope 2 uses string, not unicode yet
    name = str(name)
    if not override_existing_protection and \
            ('%s__roles__' % name) in dir(klass):
        # There is already a declaration for this name from a base class.
        return
    if permission_id == CheckerPublicId or permission_id is CheckerPublic:
        # Sometimes, we already get a processed permission id, which
        # can mean that 'zope.Public' has been interchanged for the
        # CheckerPublic object
        security.declarePublic(name)  # NOQA: D001
    elif permission_id == CheckerPrivateId:
        security.declarePrivate(name)  # NOQA: D001
    else:
        permission = getUtility(IPermission, name=permission_id)
        # Zope 2 uses string, not unicode yet
        perm = str(permission.title)
        security.declareProtected(perm, name)  # NOQA: D001


def protectClass(klass, permission_id):
    """Protect the whole class with the given permission"""
    security = _getSecurity(klass)
    if permission_id == CheckerPublicId or permission_id is CheckerPublic:
        # Sometimes, we already get a processed permission id, which
        # can mean that 'zope.Public' has been interchanged for the
        # CheckerPublic object
        security.declareObjectPublic()
    elif permission_id == CheckerPrivateId:
        security.declareObjectPrivate()
    else:
        permission = getUtility(IPermission, name=permission_id)
        # Zope 2 uses string, not unicode yet
        perm = str(permission.title)
        security.declareObjectProtected(perm)


@implementer(IConfigurationContext, IPermissionDirective)
class PermissionDirective(GroupingContextDecorator):

    def __init__(self, context, id, title, description=''):
        self.context = context
        self.id, self.title, self.description = id, title, description
        self.roles = []

    def after(self):
        permission = Permission(self.id, self.title, self.description)
        utility(self.context, IPermission, permission, name=self.id)

        zope2_permission = str(self.title)
        if self.roles:
            addPermission(zope2_permission, default_roles=tuple(self.roles))
        elif self.id == CheckerPrivateId:
            addPermission(zope2_permission, default_roles=())
        else:
            addPermission(zope2_permission)


class IRoleDirective(Interface):

    name = ASCIILine()


def RoleDirective(context, name):
    permission_directive = context.context
    if name not in permission_directive.roles:
        permission_directive.roles.append(name)
